from mpi4py import MPI

import pytest
from dolfinx.fem import Function, functionspace
from dolfinx.mesh import create_unit_cube
from ufl import exp

import festim as F

my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=None)


def test_reaction_init():
    """Test that the Reaction class initialises correctly"""
    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")

    # create a product species
    product = F.Species("C")

    # create a reaction between the two species
    reaction = F.Reaction(
        reactant=[species1, species2],
        product=product,
        k_0=1.0,
        E_k=0.2,
        p_0=0.1,
        E_p=0.3,
        volume=my_vol,
    )

    # check that the attributes are set correctly
    assert reaction.reactant == [species1, species2]
    assert reaction.product == product
    assert reaction.k_0 == 1.0
    assert reaction.E_k == 0.2
    assert reaction.p_0 == 0.1
    assert reaction.E_p == 0.3


def test_reaction_repr():
    """Test that the Reaction __repr__ method returns the expected string"""

    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")

    # create a product species
    product = F.Species("C")

    # create a reaction between the two species
    reaction = F.Reaction(
        reactant=[species1, species2],
        product=product,
        k_0=1.0,
        E_k=0.2,
        p_0=0.1,
        E_p=0.3,
        volume=my_vol,
    )

    # check that the __repr__ method returns the expected string
    expected_repr = "Reaction(A + B <--> C, 1.0, 0.2, 0.1, 0.3)"
    assert repr(reaction) == expected_repr


def test_reaction_repr_2_products():
    """Test that the Reaction __repr__ method returns the expected string"""

    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")

    # create two product species
    product1 = F.Species("C")
    product2 = F.Species("D")

    # create a reaction between the two species
    reaction = F.Reaction(
        reactant=[species1, species2],
        product=[product1, product2],
        k_0=1.0,
        E_k=0.2,
        p_0=0.1,
        E_p=0.3,
        volume=my_vol,
    )

    # check that the __repr__ method returns the expected string
    expected_repr = "Reaction(A + B <--> C + D, 1.0, 0.2, 0.1, 0.3)"
    assert repr(reaction) == expected_repr


def test_reaction_repr_0_products():
    """Test that the Reaction __repr__ method returns the expected string"""

    # create two species
    species1 = F.Species("A")

    # create a reaction between the two species
    reaction = F.Reaction(
        reactant=species1,
        k_0=1.0,
        E_k=0.2,
        volume=my_vol,
    )

    # check that the __repr__ method returns the expected string
    expected_repr = "Reaction(A <--> , 1.0, 0.2, None, None)"
    assert repr(reaction) == expected_repr


def test_reaction_str():
    """Test that the Reaction __str__ method returns the expected string"""

    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")

    # create a product species
    product = F.Species("C")

    # create a reaction between the two species
    reaction = F.Reaction(
        reactant=[species1, species2],
        product=product,
        k_0=1.0,
        E_k=0.2,
        p_0=0.1,
        E_p=0.3,
        volume=my_vol,
    )

    # check that the __str__ method returns the expected string
    expected_str = "A + B <--> C"
    assert str(reaction) == expected_str


def test_reaction_str_2_products():
    """Test that the Reaction __str__ method returns the expected string when there are 2 products"""

    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")

    # create a product species
    product1 = F.Species("C")
    product2 = F.Species("D")

    # create a reaction between the two species
    reaction = F.Reaction(
        reactant=[species1, species2],
        product=[product1, product2],
        k_0=1.0,
        E_k=0.2,
        p_0=0.1,
        E_p=0.3,
        volume=my_vol,
    )

    # check that the __str__ method returns the expected string
    expected_str = "A + B <--> C + D"
    assert str(reaction) == expected_str


def test_reaction_str_no_products():
    """Test that the Reaction __str__ method returns the expected string when there are 2 products"""

    # create two species
    species1 = F.Species("A")

    # create a reaction between the two species
    reaction = F.Reaction(
        reactant=species1,
        k_0=1.0,
        E_k=0.2,
        volume=my_vol,
    )

    # check that the __str__ method returns the expected string
    expected_str = "A <--> "
    assert str(reaction) == expected_str


@pytest.mark.parametrize("temperature", [300.0, 350, 370, 500.0])
def test_reaction_reaction_term(temperature):
    """Test that the Reaction.reaction_term method returns the expected reaction term"""

    mesh = create_unit_cube(MPI.COMM_WORLD, 10, 10, 10)
    V = functionspace(mesh, ("Lagrange", 1))

    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")
    species1.solution = Function(V)
    species2.solution = Function(V)

    # create a product species
    product = F.Species("C")
    product.solution = Function(V)

    # create a reaction between the two species
    reaction = F.Reaction(
        reactant=[species1, species2],
        product=product,
        k_0=1.0,
        E_k=0.2,
        p_0=0.1,
        E_p=0.3,
        volume=my_vol,
    )

    # test the reaction term at a given temperature
    def arrhenius(pre, act, T):
        return pre * exp(-act / (F.k_B * T))

    k = arrhenius(reaction.k_0, reaction.E_k, temperature)
    p = arrhenius(reaction.p_0, reaction.E_p, temperature)

    expected_reaction_term = (
        k * (species1.solution * species2.solution) - p * product.solution
    )

    assert reaction.reaction_term(temperature) == expected_reaction_term


@pytest.mark.parametrize("temperature", [300.0, 350, 370, 500.0])
def test_reaction_reaction_term_no_products(temperature):
    mesh = create_unit_cube(MPI.COMM_WORLD, 10, 10, 10)
    V = functionspace(mesh, ("Lagrange", 1))

    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")
    species1.solution = Function(V)
    species2.solution = Function(V)

    # create a reaction between the two species
    reaction = F.Reaction(
        reactant=[species1, species2],
        k_0=1.0,
        E_k=0.2,
        volume=my_vol,
    )

    # test the reaction term at a given temperature
    def arrhenius(pre, act, T):
        return pre * exp(-act / (F.k_B * T))

    k = arrhenius(reaction.k_0, reaction.E_k, temperature)

    expected_reaction_term = k * (species1.solution * species2.solution)

    assert reaction.reaction_term(temperature) == expected_reaction_term


@pytest.mark.parametrize("temperature", [300.0, 350, 370, 500.0])
def test_reaction_reaction_term_2_products(temperature):
    """Test that the Reaction.reaction_term method returns the expected reaction term with two products"""

    mesh = create_unit_cube(MPI.COMM_WORLD, 10, 10, 10)
    V = functionspace(mesh, ("Lagrange", 1))

    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")
    species1.solution = Function(V)
    species2.solution = Function(V)

    # create a product species
    product1 = F.Species("C")
    product2 = F.Species("D")
    product1.solution = Function(V)
    product2.solution = Function(V)

    # create a reaction between the two species
    reaction = F.Reaction(
        reactant=[species1, species2],
        product=[product1, product2],
        k_0=1.0,
        E_k=0.2,
        p_0=0.1,
        E_p=0.3,
        volume=my_vol,
    )

    # test the reaction term at a given temperature
    def arrhenius(pre, act, T):
        return pre * exp(-act / (F.k_B * T))

    k = arrhenius(reaction.k_0, reaction.E_k, temperature)
    p = arrhenius(reaction.p_0, reaction.E_p, temperature)

    product_of_products = product1.solution * product2.solution
    expected_reaction_term = (
        k * (species1.solution * species2.solution) - p * product_of_products
    )
    assert reaction.reaction_term(temperature) == expected_reaction_term


def test_reactant_setter_raises_error_with_zero_length_list():
    """Test a value error is raised when the first reactant is given a wrong type."""
    with pytest.raises(
        ValueError,
        match="reactant must be an entry of one or more species objects, not an empty list.",
    ):
        F.Reaction(
            reactant=[],
            k_0=1,
            E_k=0.1,
            p_0=2,
            E_p=0.2,
            volume=my_vol,
        )


def test_reactant_setter_raises_error_with_wrong_type():
    """Test a type error is raised when the first reactant is given a wrong type."""
    with pytest.raises(
        TypeError,
        match="reactant must be an F.Species or F.ImplicitSpecies, not <class 'str'>",
    ):
        F.Reaction(
            reactant=["A", F.Species("B")],
            product=F.Species("C"),
            k_0=1,
            E_k=0.1,
            p_0=2,
            E_p=0.2,
            volume=my_vol,
        )


def test_product_setter_raise_error_p_0_no_product():
    with pytest.raises(
        ValueError,
        match="p_0 must be None, not 2 when no products are present.",
    ):
        reaction = F.Reaction(
            reactant=[F.Species("A")],
            k_0=1,
            E_k=0.1,
            p_0=2,
            volume=my_vol,
        )
        reaction.reaction_term(temperature=500)


def test_no_E_p_with_product():
    with pytest.raises(
        ValueError,
        match="E_p cannot be None when reaction products are present.",
    ):
        reaction = F.Reaction(
            reactant=[F.Species("A")],
            product=[F.Species("C")],
            k_0=1,
            E_k=0.1,
            p_0=0.1,
            volume=my_vol,
        )
        reaction.reaction_term(temperature=500)


def test_no_p_0_with_product():
    with pytest.raises(
        ValueError,
        match="p_0 cannot be None when reaction products are present.",
    ):
        reaction = F.Reaction(
            reactant=[F.Species("A")],
            product=[F.Species("C")],
            k_0=1,
            E_k=0.1,
            E_p=1,
            volume=my_vol,
        )
        reaction.reaction_term(temperature=500)


def test_product_setter_raise_error_E_p_no_product():
    with pytest.raises(
        ValueError,
        match="E_p must be None, not 2 when no products are present.",
    ):
        reaction = F.Reaction(
            reactant=[F.Species("A")],
            k_0=1,
            E_k=0.1,
            E_p=2,
            volume=my_vol,
        )
        reaction.reaction_term(temperature=500)



import numpy as np

import festim as F

my_model = F.HydrogenTransportProblem()

# -------- Mesh --------- #

L = 5e-6
vertices = np.linspace(0, L, num=2000)
my_model.mesh = F.Mesh1D(vertices)


# -------- Materials and subdomains --------- #

w_atom_density = 6.3e28  # atom/m3

tungsten = F.Material(D_0=4.1e-7, E_D=0.39, name="tungsten")

my_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, L], material=tungsten)
left_surface = F.SurfaceSubdomain1D(id=1, x=0)
right_surface = F.SurfaceSubdomain1D(id=2, x=L)

my_model.subdomains = [
    my_subdomain,
    left_surface,
    right_surface,
]

# -------- Hydrogen species and reactions --------- #

mobile_H = F.Species("H")
mobile_D = F.Species("D")


trapped_H1 = F.Species("trapped_H1", mobile=False)
trapped_D1 = F.Species("trapped_D1", mobile=False)
trapped_H2 = F.Species("trapped_H2", mobile=False)
trapped_D2 = F.Species("trapped_D2", mobile=False)
trapped_HD = F.Species("trapped_HD", mobile=False)

empty_trap = F.ImplicitSpecies(
    n=1e21,
    others=[trapped_H1, trapped_D1, trapped_H2, trapped_HD, trapped_D2],
    name="empty_trap",
)


my_model.species = [
    mobile_H,
    mobile_D,
    trapped_H1,
    trapped_D1,
    trapped_H2,
    trapped_HD,
    trapped_D2,
]

my_model.reactions = [
    F.Reaction(
        k_0=4.1e-7 / (1.1e-10**2 * 6 * w_atom_density),
        E_k=0.39,
        p_0=1e13,
        E_p=1.2,
        reactant=[mobile_H, empty_trap],
        product=trapped_H1,
    ),
    F.Reaction(
        k_0=4.1e-7 / (1.1e-10**2 * 6 * w_atom_density),
        E_k=0.39,
        p_0=1e13,
        E_p=1.0,
        reactant=[mobile_H, trapped_H1],
        product=trapped_H2,
    ),
    F.Reaction(
        k_0=4.1e-7 / (1.1e-10**2 * 6 * w_atom_density),
        E_k=0.39,
        p_0=1e13,
        E_p=1.2,
        reactant=[mobile_D, empty_trap],
        product=trapped_D1,
    ),
    F.Reaction(
        k_0=4.1e-7 / (1.1e-10**2 * 6 * w_atom_density),
        E_k=0.39,
        p_0=1e13,
        E_p=1.2,
        reactant=[mobile_D, trapped_D1],
        product=trapped_D2,
    ),
    F.Reaction(
        k_0=4.1e-7 / (1.1e-10**2 * 6 * w_atom_density),
        E_k=0.39,
        p_0=1e13,
        E_p=1.0,
        reactant=[mobile_H, trapped_D1],
        product=trapped_HD,
    ),
    F.Reaction(
        k_0=4.1e-7 / (1.1e-10**2 * 6 * w_atom_density),
        E_k=0.39,
        p_0=1e13,
        E_p=1.0,
        reactant=[mobile_D, trapped_H1],
        product=trapped_HD,
    ),
]

# -------- Temperature --------- #

my_model.temperature = 300

# -------- Boundary conditions --------- #


my_model.boundary_conditions = [
    F.DirichletBC(subdomain=left_surface, value=1e20, species=mobile_H),
    F.DirichletBC(subdomain=right_surface, value=1e19, species=mobile_D),
    F.DirichletBC(subdomain=right_surface, value=0, species=mobile_H),
    F.DirichletBC(subdomain=left_surface, value=0, species=mobile_D),
]

# -------- Exports --------- #

left_flux = F.SurfaceFlux(field=mobile_H, surface=left_surface)
right_flux = F.SurfaceFlux(field=mobile_H, surface=right_surface)

folder = "multi_isotope_trapping_example"

my_model.exports = [
    F.XDMFExport(f"{folder}/mobile_concentration_h.xdmf", field=mobile_H),
    F.XDMFExport(f"{folder}/mobile_concentration_d.xdmf", field=mobile_D),
    F.XDMFExport(f"{folder}/trapped_concentration_h1.xdmf", field=trapped_H1),
    F.XDMFExport(f"{folder}/trapped_concentration_h2.xdmf", field=trapped_H2),
    F.XDMFExport(f"{folder}/trapped_concentration_d1.xdmf", field=trapped_D1),
    F.XDMFExport(f"{folder}/trapped_concentration_d2.xdmf", field=trapped_D2),
    F.XDMFExport(f"{folder}/trapped_concentration_hd.xdmf", field=trapped_HD),
]

# -------- Settings --------- #

my_model.settings = F.Settings(
    atol=1e-10, rtol=1e-10, max_iterations=30, final_time=3000
)

my_model.settings.stepsize = F.Stepsize(initial_value=20)

# -------- Run --------- #

my_model.initialise()

print(my_model.formulation)
# exit()
my_model.run()



import ufl

import festim as F


def henrys_law(T, H_0, E_H, pressure):
    """Applies the Henry's law to compute the concentration at the boundary"""
    H = H_0 * ufl.exp(-E_H / F.k_B / T)
    return H * pressure


class HenrysBC(F.DirichletBC):
    """
    Henrys boundary condition class

    c = H * pressure
    H = H_0 * exp(-E_H / k_B / T)

    Args:
        subdomain (festim.Subdomain): the subdomain where the boundary
            condition is applied
        species (str): the name of the species
        H_0 (float or fem.Constant): the Henrys constant pre-exponential factor (H/m3/Pa)
        E_H (float or fem.Constant): the Henrys constant activation energy (eV)
        pressure (float or callable): the pressure at the boundary (Pa)

    Attributes:
        subdomain (festim.Subdomain): the subdomain where the boundary
            condition is applied
        value (float or fem.Constant): the value of the boundary condition
        species (festim.Species or str): the name of the species
        H_0 (float or fem.Constant): the Henrys constant pre-exponential factor (H/m3/Pa)
        E_H (float or fem.Constant): the Henrys constant activation energy (eV)
        pressure (float or callable): the pressure at the boundary (Pa)

    Usage:
        >>> from festim import HenrysBC
        >>> HenrysBC(subdomain=my_subdomain, H_0=1e-6, E_H=0.2, pressure=1e5, species="H")
        >>> HenrysBC(subdomain=my_subdomain, H_0=1e-6, E_H=0.2, pressure=lambda x: 1e5 + x[0], species="H")
        >>> HenrysBC(subdomain=my_subdomain, H_0=1e-6, E_H=0.2, pressure=lambda t: 1e5 + t, species="H")
        >>> HenrysBC(subdomain=my_subdomain, H_0=1e-6, E_H=0.2, pressure=lambda T: 1e5 + T, species="H")
        >>> HenrysBC(subdomain=my_subdomain, H_0=1e-6, E_H=0.2, pressure=lambda x, t: 1e5 + x[0] + t, species="H")
    """

    def __init__(self, subdomain, H_0, E_H, pressure, species) -> None:
        # TODO find a way to have H_0 and E_H as fem.Constant
        # maybe in create_value()
        self.H_0 = H_0
        self.E_H = E_H
        self.pressure = pressure

        value = self.create_new_value_function()

        super().__init__(value=value, species=species, subdomain=subdomain)

    def create_new_value_function(self):
        """Creates a new value function based on the pressure attribute

        Raises:
            ValueError: if the pressure function is not supported

        Returns:
            callable: the value function
        """
        if callable(self.pressure):
            arg_combinations = {
                ("x",): lambda T, x=None: henrys_law(
                    T, self.H_0, self.E_H, self.pressure(x=x)
                ),
                ("t",): lambda T, t=None: henrys_law(
                    T, self.H_0, self.E_H, self.pressure(t=t)
                ),
                ("T",): lambda T: henrys_law(T, self.H_0, self.E_H, self.pressure(T=T)),
                ("t", "x"): lambda T, x=None, t=None: henrys_law(
                    T, self.H_0, self.E_H, self.pressure(x=x, t=t)
                ),
                ("T", "x"): lambda T, x=None: henrys_law(
                    T, self.H_0, self.E_H, self.pressure(x=x, T=T)
                ),
                ("T", "t"): lambda T, t=None: henrys_law(
                    T, self.H_0, self.E_H, self.pressure(t=t, T=T)
                ),
                ("T", "t", "x"): lambda T, x=None, t=None: henrys_law(
                    T, self.H_0, self.E_H, self.pressure(x=x, t=t, T=T)
                ),
            }

            # get the arguments of the pressure function
            args = self.pressure.__code__.co_varnames
            key = tuple(sorted(args))

            # get the lambda function based on the argument combination
            if key not in arg_combinations:
                raise ValueError("pressure function not supported")

            func = arg_combinations[key]

            return func
        else:
            return lambda T: henrys_law(T, self.H_0, self.E_H, self.pressure)



import numpy as np
from dolfinx import fem

import festim as F


def test_maximum_volume_compute_1D():
    """Test that the maximum volume export computes the right value"""

    # BUILD
    L = 6
    dummy_material = F.Material(D_0=1.5, E_D=1, name="dummy")
    my_mesh = F.Mesh1D(np.linspace(0, L, 10000))
    dummy_volume = F.VolumeSubdomain1D(id=1, borders=[0, L], material=dummy_material)
    dummy_volume.locate_subdomain_entities(mesh=my_mesh.mesh)

    # give function to species
    V = fem.functionspace(my_mesh.mesh, ("Lagrange", 1))
    c = fem.Function(V)
    c.interpolate(lambda x: (x[0] - 1) ** 2 + 2)

    my_species = F.Species("H")
    my_species.solution = c

    my_export = F.MaximumVolume(field=my_species, volume=dummy_volume)

    # RUN
    my_export.compute()

    # TEST
    expected_value = 27
    computed_value = my_export.value

    assert np.isclose(computed_value, expected_value, rtol=1e-2)



from dolfinx import fem

from .volume_quantity import VolumeQuantity


class AverageVolume(VolumeQuantity):
    """Computes the average value of a field in a given volume

    Args:
        field (festim.Species): species for which the average volume is computed
        volume (festim.VolumeSubdomain): volume subdomain
        filename (str, optional): name of the file to which the average volume is exported

    Attributes:
        see `festim.VolumeQuantity`
    """

    @property
    def title(self):
        return f"Average {self.field.name} volume {self.volume.id}"

    def compute(self, dx):
        """
        Computes the average value of solution function within the defined volume
        subdomain, and appends it to the data list
        """
        self.value = fem.assemble_scalar(
            fem.form(self.field.solution * dx(self.volume.id))
        ) / fem.assemble_scalar(fem.form(1 * dx(self.volume.id)))
        self.data.append(self.value)



import ufl
from dolfinx import fem

from .surface_quantity import SurfaceQuantity


class SurfaceFlux(SurfaceQuantity):
    """Computes the flux of a field on a given surface

    Args:
        field (festim.Species): species for which the surface flux is computed
        surface (festim.SurfaceSubdomain1D): surface subdomain
        filename (str, optional): name of the file to which the surface flux is exported

    Attributes:
        see `festim.SurfaceQuantity`
    """

    @property
    def title(self):
        return f"{self.field.name} flux surface {self.surface.id}"

    def compute(self, ds):
        """Computes the value of the flux at the surface

        Args:
            ds (ufl.Measure): surface measure of the model
        """

        # obtain mesh normal from field
        # if case multispecies, solution is an index, use sub_function_space
        if isinstance(self.field.solution, ufl.indexed.Indexed):
            mesh = self.field.sub_function_space.mesh
        else:
            mesh = self.field.solution.function_space.mesh
        n = ufl.FacetNormal(mesh)

        self.value = fem.assemble_scalar(
            fem.form(
                -self.D
                * ufl.dot(ufl.grad(self.field.solution), n)
                * ds(self.surface.id)
            )
        )
        self.data.append(self.value)



import numpy as np

from festim.exports import VolumeQuantity


class MinimumVolume(VolumeQuantity):
    """Computes the minimum value of a field in a given volume

    Args:
        field (festim.Species): species for which the minimum volume is computed
        volume (festim.VolumeSubdomain): volume subdomain
        filename (str, optional): name of the file to which the minimum volume is exported

    Attributes:
        see `festim.VolumeQuantity`
    """

    @property
    def title(self):
        return f"Minimum {self.field.name} volume {self.volume.id}"

    def compute(self):
        """
        Computes the minimum value of solution function within the defined volume
        subdomain, and appends it to the data list
        """
        solution = self.field.solution
        indices = self.volume.locate_subdomain_entities(solution.function_space.mesh)
        # FIXME: np.min/np.max is not parallel safe (unique value per process)
        # Needs to use a reduction operation (MPI.comm.allreduce(..., op=...))
        self.value = np.min(self.field.solution.x.array[indices])
        self.data.append(self.value)



from .dirichlet_bc import FixedConcentrationBC, FixedTemperatureBC
from .flux_bc import HeatFluxBC, ParticleFluxBC
from .surface_reaction import SurfaceReactionBC

__all__ = [
    "FixedConcentrationBC",
    "FixedTemperatureBC",
    "ParticleFluxBC",
    "SurfaceReactionBC",
    "HeatFluxBC",
]



import matplotlib.pyplot as plt
import numpy as np

times = np.genfromtxt("times_tds.txt")
flux = np.genfromtxt("outgassing_flux_tds.txt")
flux = np.abs(flux)

plt.figure()
plt.plot(times, flux)
plt.xlim(450, 500)
# plt.ylim(top=1e18)
plt.ylim(bottom=0, top=1e19)
# plt.ylim(bottom=1.25e18, top=0.6e19)
plt.ylabel(r"Desorption flux (m$^{-2}$ s$^{-1}$)")
plt.xlabel(r"Time (s)")

plt.ylabel(r"Desorption flux (m$^{-2}$ s$^{-1}$)")
plt.xlabel(r"Time (s)")
plt.show()



import os
from pathlib import Path

import mpi4py.MPI as MPI

import numpy as np
import pytest
from dolfinx import fem, mesh

import festim as F


def test_init():
    """Tests the initialisation of XDMFExport"""
    species = F.Species("H")
    my_export = F.XDMFExport(filename="my_export.xdmf", field=species)

    assert my_export.filename == Path("my_export.xdmf")
    assert my_export.field == [species]


def test_write(tmp_path):
    """Tests the write method of XDMFExport creates a file"""
    species = F.Species("H")
    filename = os.path.join(tmp_path, "test.xdmf")
    my_export = F.XDMFExport(filename=filename, field=species)
    my_export.define_writer(MPI.COMM_WORLD)
    domain = mesh.create_unit_cube(MPI.COMM_WORLD, 10, 10, 10)
    V = fem.functionspace(domain, ("Lagrange", 1))
    u = fem.Function(V)

    species.post_processing_solution = u

    for t in [0, 1, 2, 3]:
        my_export.write(t=t)

    assert os.path.exists(filename)


def test_integration_with_HTransportProblem(tmp_path):
    """Tests that XDMFExport can be used in conjunction with HTransportProblem"""
    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh1D(vertices=np.linspace(0, 1))
    my_mat = F.Material(D_0=1.9e-7, E_D=0.2, name="my_mat")
    my_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    my_model.subdomains = [my_subdomain]
    my_model.temperature = 500.0
    my_model.species = [F.Species("H")]
    filename = os.path.join(tmp_path, "test.xdmf")
    my_model.exports = [F.XDMFExport(filename=filename, field=my_model.species)]

    my_model.settings = F.Settings(atol=1, rtol=0.1, final_time=1)
    my_model.settings.stepsize = F.Stepsize(initial_value=0.5)
    my_model.initialise()
    my_model.run()

    # checks that filename exists
    assert os.path.exists(filename)


def test_field_attribute_is_always_list():
    """Test that the field attribute is always a list"""
    my_export = F.XDMFExport("my_export.xdmf", field=F.Species("H"))
    assert isinstance(my_export.field, list)

    my_export = F.XDMFExport("my_export.xdmf", field=[F.Species("H")])
    assert isinstance(my_export.field, list)


def test_vtx_suffix_converter(tmpdir):
    filename = str(tmpdir.join("my_export.txt"))
    my_export = F.XDMFExport(filename, field=[])
    assert my_export.filename.suffix == ".xdmf"


@pytest.mark.parametrize("field", [["H", 2], 1, [F.Species("H"), 1]])
def test_field_attribute_raises_error_when_invalid_type(field):
    """
    Test that the field attribute raises an error if the type
    is not festim.Species or list.
    """
    with pytest.raises(TypeError):
        F.XDMFExport("my_export.xdmf", field=field)


def test_filename_raises_error_when_wrong_type():
    """Test that the filename attribute raises an error if the file is not str"""
    with pytest.raises(TypeError):
        F.XDMFExport(1, field=[F.Species("H")])



import mpi4py.MPI as MPI

import dolfinx.mesh
import numpy as np
import pytest
import tqdm.autonotebook
import ufl
from dolfinx import default_scalar_type, fem, nls

import festim as F

test_mesh = F.Mesh1D(vertices=np.array([0.0, 1.0, 2.0, 3.0, 4.0]))
x = ufl.SpatialCoordinate(test_mesh.mesh)
dummy_mat = F.Material(D_0=1, E_D=1, name="dummy_mat")


# TODO test all the methods in the class
@pytest.mark.parametrize(
    "value",
    [
        1,
        fem.Constant(test_mesh.mesh, default_scalar_type(1.0)),
        1.0,
        "coucou",
        lambda x: 2 * x[0],
    ],
)
def test_temperature_setter_type(value):
    """Test that the temperature type is correctly set"""
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
    )

    if not isinstance(value, (fem.Constant, int, float)):
        if callable(value):
            my_model.temperature = value
        else:
            with pytest.raises(TypeError):
                my_model.temperature = value


@pytest.mark.parametrize(
    "input, expected_value",
    [
        (1.0, False),
        (1, False),
        (lambda t: t, True),
        (lambda t: 1.0 + t, True),
        (lambda x: 1.0 + x[0], False),
        (lambda x, t: 1.0 + x[0] + t, True),
        (lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0), True),
    ],
)
def test_time_dependent_temperature_attribute(input, expected_value):
    """Test that the temperature_time_dependent attribute is correctly set"""

    my_model = F.HydrogenTransportProblem()
    my_model.temperature = input

    assert my_model.temperature_time_dependent is expected_value


def test_define_temperature_value_error_raised():
    """Test that a ValueError is raised when the temperature is None"""

    # BUILD
    my_model = F.HydrogenTransportProblem(mesh=test_mesh)

    my_model.temperature = None

    # TEST
    with pytest.raises(
        ValueError, match="the temperature attribute needs to be defined"
    ):
        my_model.define_temperature()


@pytest.mark.parametrize(
    "input, expected_type",
    [
        (1.0, fem.Constant),
        (1, fem.Constant),
        (fem.Constant(test_mesh.mesh, default_scalar_type(1.0)), fem.Constant),
        (lambda t: t, fem.Constant),
        (lambda t: 1.0 + t, fem.Constant),
        (lambda x: 1.0 + x[0], fem.Function),
        (lambda x, t: 1.0 + x[0] + t, fem.Function),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0),
            fem.Function,
        ),
        (lambda t: 100.0 if t < 1 else 0.0, fem.Constant),
    ],
)
def test_define_temperature(input, expected_type):
    """Test that the define_temperature method correctly sets the
    temperature_fenics attribute to either a fem.Constant or a
    fem.Function depending on the type of input"""

    # BUILD
    my_model = F.HydrogenTransportProblem(mesh=test_mesh)
    my_model.t = fem.Constant(test_mesh.mesh, 0.0)

    my_model.temperature = input

    # RUN
    my_model.define_temperature()

    # TEST
    assert isinstance(my_model.temperature_fenics, expected_type)


@pytest.mark.parametrize(
    "input",
    [
        lambda t: ufl.conditional(ufl.lt(t, 1.0), 1, 2),
        lambda t: 1 + ufl.conditional(ufl.lt(t, 1.0), 1, 2.0),
        lambda t: 2 * ufl.conditional(ufl.lt(t, 1.0), 1, 2.0),
        lambda t: 2 / ufl.conditional(ufl.lt(t, 1.0), 1, 2.0),
    ],
)
def test_define_temperature_error_if_ufl_conditional_t_only(input):
    """Test that a ValueError is raised when the temperature attribute is a callable
    of t only and contains a ufl conditional"""
    my_model = F.HydrogenTransportProblem(mesh=test_mesh)
    my_model.t = fem.Constant(test_mesh.mesh, 0.0)

    my_model.temperature = input

    with pytest.raises(
        ValueError,
        match="self.temperature should return a float or an int, not ",
    ):
        my_model.define_temperature()


def test_iterate():
    """Test that the iterate method updates the solution and time correctly"""
    # BUILD
    my_model = F.HydrogenTransportProblem()

    my_model.settings = F.Settings(atol=1e-6, rtol=1e-6, final_time=10)
    my_model.settings.stepsize = 2.0

    my_model.progress_bar = tqdm.autonotebook.tqdm(
        desc="Solving H transport problem",
        total=my_model.settings.final_time,
        unit_scale=True,
    )

    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)

    V = fem.functionspace(mesh, ("Lagrange", 1))
    my_model.u = fem.Function(V)
    my_model.u_n = fem.Function(V)
    my_model.dt = fem.Constant(mesh, 2.0)
    v = ufl.TestFunction(V)

    source_value = 2.0
    form = (
        my_model.u - my_model.u_n
    ) / my_model.dt * v * ufl.dx - source_value * v * ufl.dx

    problem = fem.petsc.NonlinearProblem(form, my_model.u, bcs=[])
    my_model.solver = nls.petsc.NewtonSolver(MPI.COMM_WORLD, problem)

    my_model.t = fem.Constant(mesh, 0.0)

    for i in range(10):
        # RUN
        my_model.iterate()

        # TEST

        # check that t evolves
        expected_t_value = (i + 1) * float(my_model.dt)
        assert np.isclose(float(my_model.t), expected_t_value)

        # check that u and u_n are updated
        expected_u_value = (i + 1) * float(my_model.dt) * source_value
        assert np.all(np.isclose(my_model.u.x.array, expected_u_value))


@pytest.mark.parametrize(
    "T_function, expected_values",
    [
        (lambda t: t, [1.0, 2.0, 3.0]),
        (lambda t: 1.0 + t, [2.0, 3.0, 4.0]),
        (lambda x, t: 1.0 + x[0] + t, [6.0, 7.0, 8.0]),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.5), 100.0 + x[0], 0.0),
            [104.0, 0.0, 0.0],
        ),
    ],
)
def test_update_time_dependent_values_temperature(T_function, expected_values):
    """Test that different time-dependent callable functions for the
    temperature are updated at each time step and match an expected value"""

    # BUILD
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
    )
    my_model.t = fem.Constant(my_model.mesh.mesh, 0.0)
    dt = fem.Constant(test_mesh.mesh, 1.0)

    my_model.temperature = T_function

    my_model.define_temperature()

    for i in range(3):
        # RUN
        my_model.t.value += dt.value
        my_model.update_time_dependent_values()

        # TEST
        if isinstance(my_model.temperature_fenics, fem.Constant):
            computed_value = float(my_model.temperature_fenics)
            assert np.isclose(computed_value, expected_values[i])


def test_initialise_exports_find_species_with_one_field():
    """Test that a species can be found from the model species if given as a string"""

    # BUILD
    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh1D(vertices=np.array([0.0, 1.0, 2.0, 3.0, 4.0])),
        species=[F.Species("H"), F.Species("D")],
        temperature=1,
    )
    surf = F.SurfaceSubdomain1D(id=1, x=1)

    my_model.exports = [F.SurfaceFlux(field="J", surface=surf)]
    my_model.define_function_spaces()

    # TEST
    with pytest.raises(ValueError, match="Species J not found in list of species"):
        my_model.initialise_exports()


def test_define_D_global_different_temperatures():
    """Test that the D_global object is correctly defined when the temperature
    is different in the volume subdomains"""
    D_0, E_D = 1.5, 0.1
    my_mat = F.Material(D_0=D_0, E_D=E_D, name="my_mat")
    surf = F.SurfaceSubdomain1D(id=1, x=0)
    H = F.Species("H")

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh1D(np.linspace(0, 4, num=101)),
        subdomains=[
            F.VolumeSubdomain1D(id=1, borders=[0, 2], material=my_mat),
            F.VolumeSubdomain1D(id=2, borders=[2, 4], material=my_mat),
        ],
        species=[H],
        temperature=lambda x: 100.0 * x[0] + 50,
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.define_temperature()

    D_computed, D_expr = my_model.define_D_global(H)

    computed_values = [D_computed.x.array[0], D_computed.x.array[-1]]

    D_analytical_left = D_0 * np.exp(-E_D / (F.k_B * 50))
    D_analytical_right = D_0 * np.exp(-E_D / (F.k_B * 450))

    expected_values = [D_analytical_left, D_analytical_right]

    assert np.isclose(computed_values, expected_values).all()


def test_define_D_global_different_materials():
    """Test that the D_global object is correctly defined when the material
    is different in the volume subdomains"""
    D_0_left, E_D_left = 1.0, 0.1
    D_0_right, E_D_right = 2.0, 0.2
    my_mat_L = F.Material(D_0=D_0_left, E_D=E_D_left, name="my_mat_L")
    my_mat_R = F.Material(D_0=D_0_right, E_D=E_D_right, name="my_mat_R")
    H = F.Species("H")

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh1D(np.linspace(0, 4, num=101)),
        subdomains=[
            F.VolumeSubdomain1D(id=1, borders=[0, 2], material=my_mat_L),
            F.VolumeSubdomain1D(id=2, borders=[2, 4], material=my_mat_R),
        ],
        species=[H],
        temperature=500,
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.define_temperature()

    D_computed, D_expr = my_model.define_D_global(H)

    computed_values = [D_computed.x.array[0], D_computed.x.array[-1]]

    D_expected_left = D_0_left * np.exp(-E_D_left / (F.k_B * my_model.temperature))
    D_expected_right = D_0_right * np.exp(-E_D_right / (F.k_B * my_model.temperature))

    expected_values = [D_expected_left, D_expected_right]

    assert np.isclose(computed_values, expected_values).all()


def test_initialise_exports_multiple_exports_same_species():
    """Test that the diffusion coefficient within the D_global object function is the same
    for multiple exports of the same species, and that D_global object is only
    created once per species"""

    D_0, E_D = 1.5, 0.1
    my_mat = F.Material(D_0=D_0, E_D=E_D, name="my_mat")
    surf_1 = F.SurfaceSubdomain1D(id=1, x=0)
    surf_2 = F.SurfaceSubdomain1D(id=1, x=4)
    H = F.Species("H")

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh1D(np.linspace(0, 4, num=101)),
        subdomains=[
            F.VolumeSubdomain1D(id=1, borders=[0, 2], material=my_mat),
            F.VolumeSubdomain1D(id=2, borders=[2, 4], material=my_mat),
        ],
        species=[H],
        temperature=500,
        exports=[
            F.SurfaceFlux(
                field=H,
                surface=surf_1,
            ),
            F.SurfaceFlux(
                field=H,
                surface=surf_2,
            ),
        ],
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.define_temperature()
    my_model.initialise_exports()

    Ds = [export.D for export in my_model.exports]

    assert Ds[0].x.array[0] == Ds[1].x.array[0]


def test_export_resets_quantities():
    """Test that the export.data and export.t are correctly reset every time a simulation is initiated."""
    my_mat = F.Material(D_0=1, E_D=0)
    H = F.Species("H")
    surf = F.SurfaceSubdomain1D(id=1, x=4)

    my_export = F.SurfaceFlux(
        field=H,
        surface=surf,
    )

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh1D([0, 1, 2, 3, 4]),
        subdomains=[F.VolumeSubdomain1D(id=1, borders=[0, 4], material=my_mat), surf],
        species=[H],
        temperature=400,
        exports=[my_export],
    )

    my_model.settings = F.Settings(atol=1e-15, rtol=1e-15, transient=False)

    for i in range(3):
        my_model.initialise()
        my_model.run()

        assert my_model.exports[0].t == [0.0]
        assert my_model.exports[0].data == [0.0]


def test_define_D_global_multispecies():
    """Test that the D_global object is correctly defined when there are multiple
    species in one subdomain"""
    A = F.Species("A")
    B = F.Species("B")

    D_0_A, D_0_B = 1.0, 2.0
    E_D_A, E_D_B = 0.1, 0.2

    my_mat = F.Material(
        D_0={A: D_0_A, B: D_0_B}, E_D={A: E_D_A, B: E_D_B}, name="my_mat"
    )
    surf = F.SurfaceSubdomain1D(id=1, x=1)

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh1D(np.linspace(0, 1, num=101)),
        subdomains=[
            F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat),
        ],
        species=[F.Species("A"), F.Species("B")],
        temperature=500,
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.define_temperature()

    D_A_computed, D_A_expr = my_model.define_D_global(A)
    D_B_computed, D_B_expr = my_model.define_D_global(B)

    computed_values = [D_A_computed.x.array[-1], D_B_computed.x.array[-1]]

    D_expected_A = D_0_A * np.exp(-E_D_A / (F.k_B * my_model.temperature))
    D_expected_B = D_0_B * np.exp(-E_D_B / (F.k_B * my_model.temperature))

    expected_values = [D_expected_A, D_expected_B]

    assert np.isclose(computed_values, expected_values).all()


def test_post_processing_update_D_global():
    """Test that the D_global object is updated at each time
    step when temperture is time dependent"""
    my_mesh = F.Mesh1D(np.linspace(0, 1, num=11))
    my_mat = F.Material(D_0=1.5, E_D=0.1, name="my_mat")
    surf = F.SurfaceSubdomain1D(id=1, x=1)

    # create species and interpolate solution
    H = F.Species("H")
    V = fem.functionspace(my_mesh.mesh, ("Lagrange", 1))
    u = fem.Function(V)
    u.interpolate(lambda x: 2 * x[0] ** 2 + 1)
    H.solution = u

    my_export = F.SurfaceFlux(
        field=H,
        surface=surf,
    )

    # Build the model
    my_model = F.HydrogenTransportProblem(
        mesh=my_mesh,
        subdomains=[
            F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat),
            surf,
        ],
        species=[H],
        temperature=lambda t: 500 * t,
        exports=[my_export],
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.t = fem.Constant(my_model.mesh.mesh, 1.0)
    my_model.define_temperature()
    my_model.initialise_exports()

    # RUN
    my_model.post_processing()
    value_t_1 = my_export.D.x.array[-1]

    my_model.t = fem.Constant(my_model.mesh.mesh, 2.0)
    my_model.update_time_dependent_values()
    my_model.post_processing()
    value_t_2 = my_export.D.x.array[-1]

    # TEST
    assert value_t_1 != value_t_2


def test_post_processing_update_D_global_2():
    """Test that the D_global object is updated at each time
    step when temperture is time dependent"""
    my_mesh = F.Mesh1D(np.linspace(0, 1, num=11))
    my_mat = F.Material(D_0=1.5, E_D=0.1, name="my_mat")
    surf = F.SurfaceSubdomain1D(id=1, x=1)

    # create species and interpolate solution
    H = F.Species("H")
    V = fem.functionspace(my_mesh.mesh, ("Lagrange", 1))
    u = fem.Function(V)
    u.interpolate(lambda x: x[0] ** 2 + 100)
    H.solution = u

    my_export = F.MaximumSurface(
        field=H,
        surface=surf,
    )

    # Build the model
    my_model = F.HydrogenTransportProblem(
        mesh=my_mesh,
        subdomains=[
            F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat),
            surf,
        ],
        species=[H],
        temperature=lambda t: 500 * t,
        exports=[my_export],
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.t = fem.Constant(my_model.mesh.mesh, 1.0)
    my_model.define_temperature()
    my_model.initialise_exports()

    # RUN
    my_model.post_processing()
    my_value = my_export.D.x.array[-1]
    assert isinstance(my_value, float)


def test_post_processing_update_D_global_volume_1():
    """Test that the D_global object is updated at each time
    step when temperture is time dependent"""
    my_mesh = F.Mesh1D(np.linspace(0, 1, num=11))
    my_mat = F.Material(D_0=1.5, E_D=0.1, name="my_mat")
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)

    # create species and interpolate solution
    H = F.Species("H")
    V = fem.functionspace(my_mesh.mesh, ("Lagrange", 1))
    u = fem.Function(V)
    u.interpolate(lambda x: x[0] ** 2 + 100)
    H.solution = u

    my_export = F.AverageVolume(field=H, volume=my_vol)

    # Build the model
    my_model = F.HydrogenTransportProblem(
        mesh=my_mesh,
        subdomains=[my_vol],
        species=[H],
        temperature=lambda t: 500 * t,
        exports=[my_export],
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.t = fem.Constant(my_model.mesh.mesh, 1.0)
    my_model.define_temperature()
    my_model.initialise_exports()

    # RUN
    my_model.post_processing()
    my_value = my_model.exports[0].data[-1]
    assert isinstance(my_value, float)


def test_post_processing_update_D_global_volume_2():
    """Test that the D_global object is updated at each time
    step when temperture is time dependent"""
    my_mesh = F.Mesh1D(np.linspace(0, 1, num=11))
    my_mat = F.Material(D_0=1.5, E_D=0.1, name="my_mat")
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)

    # create species and interpolate solution
    H = F.Species("H")
    V = fem.functionspace(my_mesh.mesh, ("Lagrange", 1))
    u = fem.Function(V)
    u.interpolate(lambda x: x[0] ** 2 + 100)
    H.solution = u

    my_export = F.MaximumVolume(field=H, volume=my_vol)

    # Build the model
    my_model = F.HydrogenTransportProblem(
        mesh=my_mesh,
        subdomains=[my_vol],
        species=[H],
        temperature=lambda t: 500 * t,
        exports=[my_export],
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.t = fem.Constant(my_model.mesh.mesh, 1.0)
    my_model.define_temperature()
    my_model.initialise_exports()

    # RUN
    my_model.post_processing()
    my_value = my_model.exports[0].data[-1]
    assert isinstance(my_value, float)


@pytest.mark.parametrize(
    "temperature_value, bc_value, expected_values",
    [
        (5, 1.0, [1.0, 1.0, 1.0]),
        (lambda t: t + 1, lambda T: T, [2.0, 3.0, 4.0]),
        (lambda x: 1 + x[0], lambda T: 2 * T, [10.0, 10.0, 10.0]),
        (lambda x, t: t + x[0], lambda T: 0.5 * T, [2.5, 3.0, 3.5]),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 2), 3 + x[0], 0.0),
            lambda T: T,
            [7.0, 0.0, 0.0],
        ),
    ],
)
def test_update_time_dependent_bcs_with_time_dependent_temperature(
    temperature_value, bc_value, expected_values
):
    """Test that temperature dependent bcs are updated at each time step when the
    temperature is time dependent, and match an expected value"""

    # BUILD
    H = F.Species("H")
    volume_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    surface_subdomain = F.SurfaceSubdomain1D(id=1, x=4)
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        species=[H],
        subdomains=[volume_subdomain, surface_subdomain],
    )
    my_model.t = fem.Constant(my_model.mesh.mesh, 0.0)
    dt = fem.Constant(test_mesh.mesh, 1.0)

    my_model.temperature = temperature_value
    my_bc = F.DirichletBC(species=H, value=bc_value, subdomain=surface_subdomain)
    my_model.boundary_conditions = [my_bc]

    my_model.define_temperature()
    my_model.define_function_spaces()
    my_model.assign_functions_to_species()
    my_model.define_meshtags_and_measures()
    my_model.create_dirichletbc_form(my_bc)

    for i in range(3):
        # RUN
        my_model.t.value += dt.value
        my_model.update_time_dependent_values()

        # TEST
        if isinstance(my_model.boundary_conditions[0].value_fenics, fem.Constant):
            computed_value = float(my_model.boundary_conditions[0].value_fenics)
            assert np.isclose(computed_value, expected_values[i])


@pytest.mark.parametrize(
    "source_value, expected_values",
    [
        (lambda t: t, [1.0, 2.0, 3.0]),
        (lambda t: 1.0 + t, [2.0, 3.0, 4.0]),
        (lambda x, t: 1.0 + x[0] + t, [6.0, 7.0, 8.0]),
        (lambda T, t: T + 2 * t, [12.0, 14.0, 16.0]),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.5), 100.0 + x[0], 0.0),
            [104.0, 0.0, 0.0],
        ),
    ],
)
def test_update_time_dependent_values_source(source_value, expected_values):
    """Test that time dependent sources are updated at each time step,
    and match an expected value"""
    # BUILD
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    H = F.Species("H")
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh, temperature=10, subdomains=[my_vol], species=[H]
    )
    my_model.t = fem.Constant(my_model.mesh.mesh, 0.0)
    dt = fem.Constant(test_mesh.mesh, 1.0)

    my_source = F.ParticleSource(value=source_value, volume=my_vol, species=H)
    my_model.sources = [my_source]

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.assign_functions_to_species()
    my_model.define_temperature()
    my_model.create_source_values_fenics()

    for i in range(3):
        # RUN
        my_model.t.value += dt.value
        my_model.update_time_dependent_values()

        # TEST
        if isinstance(my_model.sources[0].value_fenics, fem.Constant):
            computed_value = float(my_model.sources[0].value_fenics)
            assert np.isclose(computed_value, expected_values[i])


@pytest.mark.parametrize(
    "temperature_value, source_value, expected_values",
    [
        (5, 1.0, [1.0, 1.0, 1.0]),
        (lambda t: t + 1, lambda T: T, [2.0, 3.0, 4.0]),
        (lambda x: 1 + x[0], lambda T: 2 * T, [10.0, 10.0, 10.0]),
        (lambda x, t: t + x[0], lambda T: 0.5 * T, [2.5, 3.0, 3.5]),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 2), 3 + x[0], 0.0),
            lambda T: T,
            [7.0, 0.0, 0.0],
        ),
    ],
)
def test_update_sources_with_time_dependent_temperature(
    temperature_value, source_value, expected_values
):
    """Test that temperature dependent source terms are updated at each time step
    when the temperature is time dependent, and match an expected value"""

    # BUILD
    H = F.Species("H")
    volume_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    surface_subdomain = F.SurfaceSubdomain1D(id=1, x=4)
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        species=[H],
        subdomains=[volume_subdomain, surface_subdomain],
        temperature=temperature_value,
    )
    my_model.t = fem.Constant(my_model.mesh.mesh, 0.0)
    dt = fem.Constant(test_mesh.mesh, 1.0)

    my_model.sources = [
        F.ParticleSource(value=source_value, volume=volume_subdomain, species=H)
    ]

    my_model.define_temperature()
    my_model.define_function_spaces()
    my_model.assign_functions_to_species()
    my_model.define_meshtags_and_measures()
    my_model.create_source_values_fenics()

    for i in range(3):
        # RUN
        my_model.t.value += dt.value
        my_model.update_time_dependent_values()

        # TEST
        if isinstance(my_model.sources[0].value_fenics, fem.Constant):
            computed_value = float(my_model.sources[0].value_fenics)
            assert np.isclose(computed_value, expected_values[i])


def test_create_source_values_fenics_multispecies():
    """Test that the define_sources method correctly sets the value_fenics attribute in
    a multispecies case"""
    # BUILD
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    H, D = F.Species("H"), F.Species("D")
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        temperature=10,
        subdomains=[my_vol],
        species=[H, D],
    )
    my_model.t = fem.Constant(my_model.mesh.mesh, 4.0)

    my_source_1 = F.ParticleSource(value=lambda t: t + 1, volume=my_vol, species=H)
    my_source_2 = F.ParticleSource(value=lambda t: 2 * t + 3, volume=my_vol, species=D)
    my_model.sources = [my_source_1, my_source_2]

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.assign_functions_to_species()
    my_model.define_temperature()

    # RUN
    my_model.create_source_values_fenics()

    # TEST
    assert np.isclose(my_model.sources[0].value_fenics.value, 5)
    assert np.isclose(my_model.sources[1].value_fenics.value, 11)


# TODO replace this by a proper MMS test
def test_run_in_steady_state():
    """Test that the run method works in steady state"""
    # BUILD
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        temperature=500,
        settings=F.Settings(atol=1e-10, rtol=1e-10, transient=False),
        subdomains=[my_vol],
        species=[F.Species("H")],
    )

    my_model.initialise()

    # RUN
    my_model.run()

    # TEST
    assert my_model.t.value == 0.0


def test_species_setter():
    """Test that a TypeError is rasied when a species of type other than F.Species is
    given"""

    my_model = F.HydrogenTransportProblem()

    with pytest.raises(
        TypeError,
        match="elements of species must be of type festim.Species not <class 'int'>",
    ):
        my_model.species = [1, 2, 3]


def test_create_initial_conditions_ValueError_raised_when_not_transient():
    """Test that ValueError is raised if initial conditions are defined in
    a steady state simulation"""

    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    H = F.Species("H")
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        temperature=10,
        subdomains=[my_vol],
        species=[H],
        initial_conditions=[F.InitialCondition(value=1.0, species=H)],
        settings=F.Settings(atol=1, rtol=1, transient=False),
    )

    with pytest.raises(
        ValueError,
        match="Initial conditions can only be defined for transient simulations",
    ):
        my_model.initialise()


@pytest.mark.parametrize(
    "input_value, expected_value",
    [
        (1.0, 1.0),
        (1, 1.0),
        (lambda T: 1.0 + T, 11.0),
        (lambda x: 1.0 + x[0], 5.0),
        (lambda x, T: 1.0 + x[0] + T, 15.0),
    ],
)
def test_create_initial_conditions_expr_fenics(input_value, expected_value):
    """Test that after calling create_initial_conditions, the prev_solution
    attribute of the species has the correct value at x=4.0."""

    # BUILD
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 4], material=dummy_mat)
    H = F.Species("H")
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        temperature=10,
        subdomains=[vol_subdomain],
        species=[H],
        initial_conditions=[F.InitialCondition(value=input_value, species=H)],
        settings=F.Settings(atol=1, rtol=1, final_time=2, stepsize=1),
    )

    # RUN
    my_model.initialise()

    assert np.isclose(
        my_model.species[0].prev_solution.x.petsc_vec.array[-1],
        expected_value,
    )


def test_create_species_from_trap():
    "Test that a new species and reaction is created when a trap is given"

    # BUILD
    my_model = F.HydrogenTransportProblem(mesh=test_mesh)
    my_mobile_species = F.Species("test_mobile")
    mat = F.Material(D_0=1, E_D=1, name="mat")
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=mat)
    my_trap = F.Trap(
        name="test_trap",
        mobile_species=my_mobile_species,
        k_0=1,
        E_k=1,
        p_0=1,
        E_p=1,
        n=1,
        volume=my_vol,
    )
    my_settings = F.Settings(atol=1, rtol=1, transient=False)
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        subdomains=[my_vol],
        species=[my_mobile_species],
        traps=[my_trap],
        temperature=100,
        settings=my_settings,
    )

    # RUN
    my_model.initialise()

    # TEST
    # test that an additional species is generated
    assert len(my_model.species) == 2
    assert isinstance(my_model.species[1], F.Species)

    assert len(my_model.reactions) == 1
    assert isinstance(my_model.reactions[0], F.Reaction)


@pytest.mark.parametrize(
    "input_value_1, input_value_2, expected_value_1, expected_value_2",
    [
        (1.0, 1.0, 1.0, 1.0),
        (1.0, 1, 1.0, 1.0),
        (1.0, lambda T: 1.0 + T, 1.0, 11.0),
        (1.0, lambda x: 1.0 + x[0], 1.0, 5.0),
        (1.0, lambda x, T: 1.0 + x[0] + T, 1.0, 15.0),
    ],
)
def test_create_initial_conditions_value_fenics_multispecies(
    input_value_1, input_value_2, expected_value_1, expected_value_2
):
    """Test that after calling create_initial_conditions, the prev_solution
    attribute of each species has the correct value at x=4.0 in a multispecies case"""

    # BUILD
    test_mesh = F.Mesh1D(vertices=np.linspace(0, 4, num=101))
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 4], material=dummy_mat)
    H, D = F.Species("H"), F.Species("D")
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        temperature=10,
        subdomains=[vol_subdomain],
        species=[H, D],
        initial_conditions=[
            F.InitialCondition(value=input_value_2, species=D),
            F.InitialCondition(value=input_value_1, species=H),
        ],
        settings=F.Settings(atol=1, rtol=1, final_time=2, stepsize=1),
    )

    # RUN
    my_model.initialise()

    # TEST
    # When in multispecies, the u and u_n x arrays are structured as follows:
    # [H, D, ..., H, D, H, D], thus the last two values are the ones we are
    # interested in

    # test value of H at x = 4.0
    assert np.isclose(my_model.u_n.x.array[-2], expected_value_1)

    # test value of D at x = 4.0
    assert np.isclose(my_model.u_n.x.array[-1], expected_value_2)


def test_adaptive_timestepping_grows():
    """Tests that the stepsize grows"""
    # BUILD
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        temperature=500,
        settings=F.Settings(atol=1e-10, rtol=1e-10, transient=True, final_time=10),
        subdomains=[my_vol],
        species=[F.Species("H")],
    )

    stepsize = F.Stepsize(initial_value=1)
    stepsize.growth_factor = 1.2
    stepsize.target_nb_iterations = 100  # force it to always grow
    my_model.settings.stepsize = stepsize

    my_model.initialise()

    my_model.progress_bar = tqdm.autonotebook.tqdm(
        desc="Solving H transport problem",
        total=my_model.settings.final_time,
        unit_scale=True,
    )

    # RUN & TEST
    previous_value = stepsize.initial_value
    while my_model.t.value < my_model.settings.final_time:
        my_model.iterate()

        # check that the current value is greater than the previous one
        assert my_model.dt.value > previous_value

        previous_value = float(my_model.dt)


def test_adaptive_timestepping_shrinks():
    """Tests that the stepsize shrinks"""
    # BUILD
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        temperature=500,
        settings=F.Settings(atol=1e-10, rtol=1e-10, transient=True, final_time=10),
        subdomains=[my_vol],
        species=[F.Species("H")],
    )

    stepsize = F.Stepsize(initial_value=1)
    stepsize.cutback_factor = 0.8
    stepsize.target_nb_iterations = -1  # force it to always shrink
    my_model.settings.stepsize = stepsize

    my_model.initialise()

    my_model.progress_bar = tqdm.autonotebook.tqdm(
        desc="Solving H transport problem",
        total=my_model.settings.final_time,
        unit_scale=True,
    )

    # RUN & TEST
    previous_value = stepsize.initial_value
    while my_model.t.value < my_model.settings.final_time and my_model.dt.value > 0.1:
        my_model.iterate()

        # check that the current value is smaller than the previous one
        assert my_model.dt.value < previous_value

        previous_value = float(my_model.dt)


@pytest.mark.parametrize(
    "attribute, value",
    [
        ("species", F.Species("test")),
        ("reactions", None),
        ("sources", None),
        ("subdomains", None),
        ("boundary_conditions", None),
        ("exports", None),
    ],
)
def test_reinstantiation_of_class(attribute, value):
    """Test that when an attribute defaults to empty list, when the class
    is reinstantiated the list is not passed to the new class object"""

    model_1 = F.HydrogenTransportProblem()
    getattr(model_1, attribute).append(value)

    model_2 = F.HydrogenTransportProblem()
    assert len(getattr(model_2, attribute)) == 0


def test_define_meshtags_and_measures_with_custom_fenics_mesh():
    """Test that the define_meshtags_and_measures method works when the mesh is
    a custom fenics mesh"""

    # BUILD
    mesh_1D = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 10)
    # 1D meshtags
    my_surface_meshtags = dolfinx.mesh.meshtags(
        mesh_1D,
        0,
        np.array([0, 10], dtype=np.int32),
        np.array([1, 2], dtype=np.int32),
    )

    num_cells = mesh_1D.topology.index_map(1).size_local
    my_volume_meshtags = dolfinx.mesh.meshtags(
        mesh_1D,
        1,
        np.arange(num_cells, dtype=np.int32),
        np.full(num_cells, 1, dtype=np.int32),
    )

    my_mesh = F.Mesh(mesh=mesh_1D)

    my_model = F.HydrogenTransportProblem(mesh=my_mesh)
    my_model.facet_meshtags = my_surface_meshtags
    my_model.volume_meshtags = my_volume_meshtags

    # TEST
    my_model.define_meshtags_and_measures()


def test_error_raised_when_custom_fenics_mesh_wrong_facet_meshtags_type():
    """Test the facet_meshtags type hinting raises error when given as wrong type"""

    # BUILD
    mesh_1D = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 10)
    my_mesh = F.Mesh(mesh=mesh_1D)
    my_model = F.HydrogenTransportProblem(mesh=my_mesh)

    # TEST
    with pytest.raises(TypeError, match="value must be of type dolfinx.mesh.MeshTags"):
        my_model.facet_meshtags = [0, 1]


def test_error_raised_when_custom_fenics_mesh_wrong_volume_meshtags_type():
    """Test the volume_meshtags type hinting raises error when given as wrong type"""

    # BUILD
    mesh_1D = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 10)
    my_mesh = F.Mesh(mesh=mesh_1D)
    my_model = F.HydrogenTransportProblem(mesh=my_mesh)

    # TEST
    with pytest.raises(TypeError, match="value must be of type dolfinx.mesh.MeshTags"):
        my_model.volume_meshtags = [0, 1]


@pytest.mark.parametrize(
    "bc_value, expected_values",
    [
        (lambda t: t, [1.0, 2.0, 3.0]),
        (lambda t: 1.0 + t, [2.0, 3.0, 4.0]),
        (lambda x, t: 1.0 + x[0] + t, [6.0, 7.0, 8.0]),
        (lambda T, t: T + 2 * t, [12.0, 14.0, 16.0]),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.5), 100.0 + x[0], 0.0),
            [104.0, 0.0, 0.0],
        ),
    ],
)
def test_update_time_dependent_values_flux(bc_value, expected_values):
    """Test that time dependent fluxes are updated at each time step,
    and match an expected value"""
    # BUILD
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    surface = F.SurfaceSubdomain1D(id=2, x=0)
    H = F.Species("H")
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        temperature=10,
        subdomains=[my_vol, surface],
        species=[H],
    )
    my_model.t = fem.Constant(my_model.mesh.mesh, 0.0)
    dt = fem.Constant(test_mesh.mesh, 1.0)

    my_bc = F.ParticleFluxBC(subdomain=surface, value=bc_value, species=H)
    my_model.boundary_conditions = [my_bc]

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.assign_functions_to_species()
    my_model.define_temperature()
    my_model.define_boundary_conditions()

    for i in range(3):
        # RUN
        my_model.t.value += dt.value
        my_model.update_time_dependent_values()

        # TEST
        if isinstance(my_model.boundary_conditions[0].value_fenics, fem.Constant):
            computed_value = float(my_model.boundary_conditions[0].value_fenics)
            assert np.isclose(computed_value, expected_values[i])


@pytest.mark.parametrize(
    "temperature_value, bc_value, expected_values",
    [
        (5, 1.0, [1.0, 1.0, 1.0]),
        (lambda t: t + 1, lambda T: T, [2.0, 3.0, 4.0]),
        (lambda x: 1 + x[0], lambda T: 2 * T, [10.0, 10.0, 10.0]),
        (lambda x, t: t + x[0], lambda T: 0.5 * T, [2.5, 3.0, 3.5]),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 2), 3 + x[0], 0.0),
            lambda T: T,
            [7.0, 0.0, 0.0],
        ),
    ],
)
def test_update_fluxes_with_time_dependent_temperature(
    temperature_value, bc_value, expected_values
):
    """Test that temperature dependent flux terms are updated at each time step
    when the temperature is time dependent, and match an expected value"""

    # BUILD
    H = F.Species("H")
    volume_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    surface_subdomain = F.SurfaceSubdomain1D(id=1, x=4)
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        species=[H],
        subdomains=[volume_subdomain, surface_subdomain],
        temperature=temperature_value,
    )
    my_model.t = fem.Constant(my_model.mesh.mesh, 0.0)
    dt = fem.Constant(test_mesh.mesh, 1.0)

    my_model.boundary_conditions = [
        F.ParticleFluxBC(subdomain=surface_subdomain, value=bc_value, species=H)
    ]

    my_model.define_temperature()
    my_model.define_function_spaces()
    my_model.assign_functions_to_species()
    my_model.define_meshtags_and_measures()
    my_model.define_boundary_conditions()

    for i in range(3):
        # RUN
        my_model.t.value += dt.value
        my_model.update_time_dependent_values()

        # TEST
        if isinstance(my_model.boundary_conditions[0].value_fenics, fem.Constant):
            computed_value = float(my_model.boundary_conditions[0].value_fenics)
            assert np.isclose(computed_value, expected_values[i])


def test_create_flux_values_fenics_multispecies():
    """Test that the create_flux_values_fenics method correctly sets the value_fenics
    attribute in a multispecies case"""
    # BUILD
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    surface = F.SurfaceSubdomain1D(id=2, x=0)
    H, D = F.Species("H"), F.Species("D")
    my_model = F.HydrogenTransportProblem(
        mesh=test_mesh,
        temperature=10,
        subdomains=[my_vol, surface],
        species=[H, D],
    )
    my_model.t = fem.Constant(my_model.mesh.mesh, 4.0)

    my_bc_1 = F.ParticleFluxBC(subdomain=surface, value=lambda t: t + 1, species=H)
    my_bc_2 = F.ParticleFluxBC(subdomain=surface, value=lambda t: 2 * t + 3, species=D)
    my_model.boundary_conditions = [my_bc_1, my_bc_2]

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()
    my_model.assign_functions_to_species()
    my_model.define_temperature()

    # RUN
    my_model.create_flux_values_fenics()

    # TEST
    assert np.isclose(my_model.boundary_conditions[0].value_fenics.value, 5)
    assert np.isclose(my_model.boundary_conditions[1].value_fenics.value, 11)



from mpi4py import MPI

import dolfinx
import numpy as np
import pytest

import festim as F

mesh = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 10, 10, 10)
V = dolfinx.fem.functionspace(mesh, ("Lagrange", 1))


def test_vtx_export_one_function(tmpdir):
    """Test can add one function to a vtx export"""
    u = dolfinx.fem.Function(V)
    sp = F.Species("H")
    sp.post_processing_solution = u
    filename = str(tmpdir.join("my_export.bp"))
    my_export = F.VTXSpeciesExport(filename, field=sp)

    functions = my_export.get_functions()
    assert len(functions) == 1
    assert functions[0] == u


def test_vtx_export_two_functions(tmpdir):
    """Test can add two functions to a vtx export"""
    u = dolfinx.fem.Function(V)
    v = dolfinx.fem.Function(V)

    sp1 = F.Species("1")
    sp2 = F.Species("2")
    sp1.post_processing_solution = u
    sp2.post_processing_solution = v
    filename = str(tmpdir.join("my_export.bp"))
    my_export = F.VTXSpeciesExport(filename, field=[sp1, sp2])

    functions = my_export.get_functions()
    assert len(functions) == 2
    assert functions[0] == u
    assert functions[1] == v


@pytest.mark.skip(reason="Not implemented")
def test_vtx_export_subdomain():
    """Test that given multiple subdomains in problem,
    only correct functions are extracted from species"""
    pass


def test_vtx_suffix_converter(tmpdir):
    filename = str(tmpdir.join("my_export.txt"))
    my_export = F.VTXSpeciesExport(filename, field=[])
    assert my_export.filename.suffix == ".bp"


def test_vtx_DG(tmpdir):
    """Test VTX export setup for DG formulation"""
    my_model = F.HTransportProblemDiscontinuous()
    my_model.mesh = F.Mesh1D(vertices=np.array([0.0, 1.0, 2.0, 3.0, 4.0]))
    my_mat = F.Material(D_0=3, E_D=2, K_S_0=1, E_K_S=0, name="mat")

    s0 = F.VolumeSubdomain1D(1, borders=[0.0, 2], material=my_mat)
    s1 = F.VolumeSubdomain1D(2, borders=[2, 4], material=my_mat)
    l0 = F.SurfaceSubdomain1D(1, x=0.0)
    l1 = F.SurfaceSubdomain1D(2, x=4.0)
    my_model.interfaces = [F.Interface(6, (s0, s1))]

    my_model.temperature = 55
    my_model.subdomains = [s0, s1, l0, l1]
    my_model.surface_to_volume = {l0: s0, l1: s1}
    # NOTE: Ask Remi why `H` has to live in both s0 and s1
    my_model.species = [
        F.Species("H", subdomains=[s0, s1]),
        F.Species("T", subdomains=[s0, s1], mobile=False),
    ]

    filename = str(tmpdir.join("my_export.txt"))
    my_export = F.VTXSpeciesExport(filename, field=my_model.species, subdomain=s0)
    assert my_export.filename.suffix == ".bp"
    my_model.exports = [my_export]
    my_model.settings = F.Settings(atol=1, rtol=0.1)
    my_model.settings.stepsize = F.Stepsize(initial_value=1)

    my_model.initialise()
    assert len(my_export.get_functions()) == 2
    assert len(my_model._vtxfiles) == 1


def test_vtx_integration_with_h_transport_problem(tmpdir):
    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh1D(vertices=np.array([0.0, 1.0, 2.0, 3.0, 4.0]))
    my_mat = F.Material(D_0=1, E_D=0, name="mat")
    my_model.subdomains = [
        F.VolumeSubdomain1D(1, borders=[0.0, 4.0], material=my_mat),
        F.SurfaceSubdomain1D(1, x=0.0),
        F.SurfaceSubdomain1D(2, x=4.0),
    ]
    my_model.species = [F.Species("H")]
    my_model.temperature = 500

    filename = str(tmpdir.join("my_export.bp"))
    my_export = F.VTXSpeciesExport(filename, field=my_model.species[0])
    my_model.exports = [my_export]
    my_model.settings = F.Settings(atol=1, rtol=0.1)
    my_model.settings.stepsize = F.Stepsize(initial_value=1)

    my_model.initialise()
    assert len(my_export.get_functions()) == 1
    assert len(my_model._vtxfiles) == 1


def test_field_attribute_is_always_list():
    """Test that the field attribute is always a list"""
    my_export = F.VTXSpeciesExport("my_export.bp", field=F.Species("H"))
    assert isinstance(my_export.field, list)

    my_export = F.VTXSpeciesExport("my_export.bp", field=[F.Species("H")])
    assert isinstance(my_export.field, list)


@pytest.mark.parametrize("field", [["H", 1], 1, [F.Species("H"), 1]])
def test_field_attribute_raises_error_when_invalid_type(field):
    """Test that the field attribute raises an error if the type is not festim.Species or list"""
    with pytest.raises(TypeError):
        F.VTXSpeciesExport("my_export.bp", field=field)


def test_filename_raises_error_when_wrong_type():
    """Test that the filename attribute raises an error if the extension is not .bp"""
    with pytest.raises(TypeError):
        F.VTXSpeciesExport(1, field=[F.Species("H")])



import dolfinx
from dolfinx import fem


def as_fenics_constant(
    value: float | int | fem.Constant, mesh: dolfinx.mesh.Mesh
) -> fem.Constant:
    """Converts a value to a dolfinx.Constant

    Args:
        value: the value to convert
        mesh: the mesh of the domiain

    Returns:
        The converted value

    Raises:
        TypeError: if the value is not a float, an int or a dolfinx.Constant
    """
    if isinstance(value, (float, int)):
        return fem.Constant(mesh, dolfinx.default_scalar_type(value))
    elif isinstance(value, fem.Constant):
        return value
    else:
        raise TypeError(
            f"Value must be a float, an int or a dolfinx.Constant, not {type(value)}"
        )



import os

import numpy as np
import pytest
import ufl
from dolfinx import fem
from dolfinx.mesh import meshtags

import festim as F


def test_surface_flux_export_compute():
    """Test that the surface flux export computes the correct value"""

    # BUILD
    L = 4.0
    D = 1.5
    my_mesh = F.Mesh1D(np.linspace(0, L, 10000))
    dummy_surface = F.SurfaceSubdomain1D(id=1, x=4)

    # define mesh ds measure
    facet_indices = np.array(
        dummy_surface.locate_boundary_facet_indices(my_mesh.mesh),
        dtype=np.int32,
    ).flatten()
    tags_facets = np.array(
        [1],
        dtype=np.int32,
    ).flatten()
    facet_meshtags = meshtags(my_mesh.mesh, 0, facet_indices, tags_facets)
    ds = ufl.Measure("ds", domain=my_mesh.mesh, subdomain_data=facet_meshtags)

    # give function to species
    V = fem.functionspace(my_mesh.mesh, ("Lagrange", 1))
    c = fem.Function(V)
    c.interpolate(lambda x: 2 * x[0] ** 2 + 1)

    my_species = F.Species("H")
    my_species.solution = c

    my_export = F.SurfaceFlux(
        field=my_species,
        surface=dummy_surface,
    )
    my_export.D = D

    # RUN
    my_export.compute(ds=ds)

    # TEST
    # flux = -D grad(c)_ \cdot n = -D dc/dx = -D * 4 * x
    expected_value = -D * 4 * dummy_surface.x
    computed_value = my_export.value

    assert np.isclose(computed_value, expected_value, rtol=1e-2)


@pytest.mark.parametrize("value", ["my_export.csv", "my_export.txt"])
def test_title_generation(tmp_path, value):
    """Test that the title is made to be written to the header in a csv or txt file"""
    my_export = F.SurfaceFlux(
        filename=os.path.join(tmp_path, f"{value}"),
        field=F.Species("TEST"),
        surface=F.SurfaceSubdomain1D(id=35, x=1),
    )
    my_export.value = 2.0
    my_export.write(0)
    title = np.genfromtxt(my_export.filename, delimiter=",", max_rows=1, dtype=str)

    expected_title = "TEST flux surface 35"

    assert title[1] == expected_title


def test_write_overwrite(tmp_path):
    """Test that the write method overwrites the file if it already exists"""
    filename = os.path.join(tmp_path, "my_export.csv")
    my_export = F.SurfaceFlux(
        filename=filename,
        field=F.Species("test"),
        surface=F.SurfaceSubdomain1D(id=1, x=0),
    )
    my_export.value = 2.0
    my_export.write(0)
    my_export.write(1)

    my_export2 = F.SurfaceFlux(
        filename=filename,
        field=F.Species("test"),
        surface=F.SurfaceSubdomain1D(id=1, x=0),
    )
    my_export2.value = 3.0
    my_export2.write(1)
    my_export2.write(2)
    my_export2.write(3)

    data = np.genfromtxt(filename, delimiter=",", names=True)
    file_length = data.size
    expected_length = 3

    assert file_length == expected_length


def test_filename_setter_raises_TypeError():
    """Test that a TypeError is raised when the filename is not a string"""

    with pytest.raises(TypeError, match="filename must be of type str"):
        F.SurfaceQuantity(
            filename=1,
            field=F.Species("test"),
            surface=F.SurfaceSubdomain1D(id=1, x=0),
        )


def test_filename_setter_raises_ValueError(tmp_path):
    """Test that a ValueError is raised when the filename does not end with .csv or .txt"""

    with pytest.raises(ValueError):
        F.SurfaceQuantity(
            filename=os.path.join(tmp_path, "my_export.xdmf"),
            field=F.Species("test"),
            surface=F.SurfaceSubdomain1D(id=1, x=0),
        )


def test_field_setter_raises_TypeError():
    """Test that a TypeError is raised when the field is not a F.Species"""

    with pytest.raises(TypeError):
        F.SurfaceQuantity(
            field=1,
            surface=F.SurfaceSubdomain1D(id=1, x=0),
        )


@pytest.mark.parametrize("value", ["my_export.csv", "my_export.txt"])
def test_writer(tmp_path, value):
    """Test that the writes values at each timestep to either a csv or txt file"""
    my_export = F.SurfaceFlux(
        filename=os.path.join(tmp_path, f"{value}"),
        field=F.Species("test"),
        surface=F.SurfaceSubdomain1D(id=1, x=0),
    )
    my_export.value = 2.0

    for i in range(10):
        my_export.write(i)
        file_length = len(np.genfromtxt(my_export.filename, delimiter=","))

        expected_length = i + 2

        assert file_length == expected_length


def test_surface_setter_raises_TypeError():
    """Test that a TypeError is raised when the surface is not a
    F.SurfaceSubdomain"""

    with pytest.raises(
        TypeError, match="surface should be an int or F.SurfaceSubdomain"
    ):
        F.SurfaceQuantity(
            field=F.Species("H"),
            surface="1",
        )



import ufl

import festim as F


def sieverts_law(T, S_0, E_S, pressure):
    """Applies the Sieverts law to compute the concentration at the boundary"""
    S = S_0 * ufl.exp(-E_S / F.k_B / T)
    return S * pressure**0.5


class SievertsBC(F.DirichletBC):
    """
    Sieverts boundary condition class

    c = S * sqrt(pressure)
    S = S_0 * exp(-E_S / k_B / T)

    Args:
        subdomain (festim.Subdomain): the subdomain where the boundary
            condition is applied
        species (str): the name of the species
        S_0 (float or fem.Constant): the Sieverts constant pre-exponential factor (H/m3/Pa0.5)
        E_S (float or fem.Constant): the Sieverts constant activation energy (eV)
        pressure (float or callable): the pressure at the boundary (Pa)

    Attributes:
        subdomain (festim.Subdomain): the subdomain where the boundary
            condition is applied
        value (float or fem.Constant): the value of the boundary condition
        species (festim.Species or str): the name of the species
        S_0 (float or fem.Constant): the Sieverts constant pre-exponential factor (H/m3/Pa0.5)
        E_S (float or fem.Constant): the Sieverts constant activation energy (eV)
        pressure (float or callable): the pressure at the boundary (Pa)

    Usage:
        >>> from festim import SievertsBC
        >>> SievertsBC(subdomain=my_subdomain, S_0=1e-6, E_S=0.2, pressure=1e5, species="H")
        >>> SievertsBC(subdomain=my_subdomain, S_0=1e-6, E_S=0.2, pressure=lambda x: 1e5 + x[0], species="H")
        >>> SievertsBC(subdomain=my_subdomain, S_0=1e-6, E_S=0.2, pressure=lambda t: 1e5 + t, species="H")
        >>> SievertsBC(subdomain=my_subdomain, S_0=1e-6, E_S=0.2, pressure=lambda T: 1e5 + T, species="H")
        >>> SievertsBC(subdomain=my_subdomain, S_0=1e-6, E_S=0.2, pressure=lambda x, t: 1e5 + x[0] + t, species="H")
    """

    def __init__(self, subdomain, S_0, E_S, pressure, species) -> None:
        # TODO find a way to have S_0 and E_S as fem.Constant
        # maybe in create_value()
        self.S_0 = S_0
        self.E_S = E_S
        self.pressure = pressure

        value = self.create_new_value_function()

        super().__init__(value=value, species=species, subdomain=subdomain)

    def create_new_value_function(self):
        """Creates a new value function based on the pressure attribute

        Raises:
            ValueError: if the pressure function is not supported

        Returns:
            callable: the value function
        """
        if callable(self.pressure):
            arg_combinations = {
                ("x",): lambda T, x=None: sieverts_law(
                    T, self.S_0, self.E_S, self.pressure(x=x)
                ),
                ("t",): lambda T, t=None: sieverts_law(
                    T, self.S_0, self.E_S, self.pressure(t=t)
                ),
                ("T",): lambda T: sieverts_law(
                    T, self.S_0, self.E_S, self.pressure(T=T)
                ),
                ("t", "x"): lambda T, x=None, t=None: sieverts_law(
                    T, self.S_0, self.E_S, self.pressure(x=x, t=t)
                ),
                ("T", "x"): lambda T, x=None: sieverts_law(
                    T, self.S_0, self.E_S, self.pressure(x=x, T=T)
                ),
                ("T", "t"): lambda T, t=None: sieverts_law(
                    T, self.S_0, self.E_S, self.pressure(t=t, T=T)
                ),
                ("T", "t", "x"): lambda T, x=None, t=None: sieverts_law(
                    T, self.S_0, self.E_S, self.pressure(x=x, t=t, T=T)
                ),
            }

            # get the arguments of the pressure function
            args = self.pressure.__code__.co_varnames
            key = tuple(sorted(args))

            # get the lambda function based on the argument combination
            if key not in arg_combinations:
                raise ValueError("pressure function not supported")

            func = arg_combinations[key]

            return func
        else:
            return lambda T: sieverts_law(T, self.S_0, self.E_S, self.pressure)



from festim.subdomain.volume_subdomain import (
    VolumeSubdomain as _VolumeSubdomain,
)
from festim.helpers import as_fenics_constant

from typing import List, Union
import ufl
from dolfinx import fem


class Species:
    """
    Hydrogen species class for H transport simulation.

    Args:
        name (str, optional): a name given to the species. Defaults to None.
        mobile (bool, optional): whether the species is mobile or not.
            Defaults to True.
        subdomain (F.VolumeSubdomain, optional): the volume subdomain where the
            species is. Defaults to None.

    Attributes:
        name (str): a name given to the species.
        mobile (bool): whether the species is mobile or not.
        solution (dolfinx.fem.Function): the solution for the current timestep
        prev_solution (dolfinx.fem.Function): the solution for the previous
            timestep
        test_function (ufl.Argument): the testfunction associated with this
            species
        sub_function_space (dolfinx.fem.function.FunctionSpaceBase): the
            subspace of the function space
        collapsed_function_space (dolfinx.fem.function.FunctionSpaceBase): the
            collapsed function space for a species in the function space. In
            case single species case, this is None.
        post_processing_solution (dolfinx.fem.Function): the solution for post
            processing
        concentration (dolfinx.fem.Function): the concentration of the species
        subdomains (F.VolumeSubdomain): the volume subdomains where the species is
        subdomain_to_solution (dict): a dictionary mapping subdomains to solutions
        subdomain_to_prev_solution (dict): a dictionary mapping subdomains to
            previous solutions
        subdomain_to_test_function (dict): a dictionary mapping subdomains to
            test functions
        subdomain_to_post_processing_solution (dict): a dictionary mapping
            subdomains to post processing solutions
        subdomain_to_collapsed_function_space (dict): a dictionary mapping
            subdomains to collapsed function spaces
        subdomain_to_function_space (dict): a dictionary mapping subdomains to
            function spaces

    Usage:
        >>> from festim import Species, HTransportProblem
        >>> species = Species(name="H")
        >>> species.name
        'H'
        >>> my_model = HTransportProblem()
        >>> my_model.species.append(species)

    """

    subdomains: list[_VolumeSubdomain] | _VolumeSubdomain
    subdomain_to_solution: dict
    subdomain_to_prev_solution: dict
    subdomain_to_test_function: dict
    subdomain_to_post_processing_solution: dict
    subdomain_to_collapsed_function_space: dict
    subdomain_to_function_space: dict

    def __init__(self, name: str = None, mobile=True, subdomains=None) -> None:
        self.name = name
        self.mobile = mobile
        self.solution = None
        self.prev_solution = None
        self.test_function = None
        self.sub_function_space = None
        self.post_processing_solution = None
        self.collapsed_function_space = None

        self.subdomains = subdomains
        self.subdomain_to_solution = {}
        self.subdomain_to_prev_solution = {}
        self.subdomain_to_test_function = {}
        self.subdomain_to_post_processing_solution = {}
        self.subdomain_to_collapsed_function_space = {}
        self.subdomain_to_function_space = {}

    def __repr__(self) -> str:
        return f"Species({self.name})"

    def __str__(self) -> str:
        return f"{self.name}"

    @property
    def concentration(self):
        return self.solution

    @property
    def legacy(self) -> bool:
        """
        Check if we are using FESTIM 1.0 implementation or FESTIM 2.0
        """
        if not self.subdomain_to_solution:
            return True
        else:
            return False


class ImplicitSpecies:
    """Implicit species class for H transport simulation.
    c = n - others

    Args:
        n (Union[float, callable]): the total concentration of the species
        others (list[Species]): the list of species from which the implicit
            species concentration is computed (c = n - others)
        name (str, optional): a name given to the species. Defaults to None.

    Attributes:
        name (str): a name given to the species.
        n (float): the total concentration of the species
        others (list[Species]): the list of species from which the implicit
            species concentration is computed (c = n - others)
        concentration (form): the concentration of the species
        value_fenics: the total concentration as a fenics object
    """

    def __init__(
        self,
        n: Union[float, callable],
        others: List[Species] = None,
        name: str = None,
    ) -> None:
        self.name = name
        self.n = n
        self.others = others

    def __repr__(self) -> str:
        return f"ImplicitSpecies({self.name}, {self.n}, {self.others})"

    def __str__(self) -> str:
        return f"{self.name}"

    @property
    def concentration(self):
        if len(self.others) > 0:
            for other in self.others:
                if other.solution is None:
                    raise ValueError(
                        f"Cannot compute concentration of {self.name} "
                        + f"because {other.name} has no solution."
                    )
        return self.value_fenics - sum([other.solution for other in self.others])

    def create_value_fenics(self, mesh, t: fem.Constant):
        """Creates the value of the density as a fenics object and sets it to
        self.value_fenics.
        If the value is a constant, it is converted to a fenics.Constant.
        If the value is a function of t, it is converted to a fenics.Constant.
        Otherwise, it is converted to a ufl Expression

        Args:
            mesh (dolfinx.mesh.Mesh) : the mesh
            t (dolfinx.fem.Constant): the time
        """
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.n, (int, float)):
            self.value_fenics = as_fenics_constant(mesh=mesh, value=self.n)

        elif isinstance(self.n, (fem.Function, ufl.core.expr.Expr)):
            self.value_fenics = self.n

        elif callable(self.n):
            arguments = self.n.__code__.co_varnames

            if "t" in arguments and "x" not in arguments and "T" not in arguments:
                # only t is an argument
                if not isinstance(self.n(t=float(t)), (float, int)):
                    raise ValueError(
                        f"self.value should return a float or an int, not {type(self.n(t=float(t)))} "
                    )
                self.value_fenics = as_fenics_constant(
                    mesh=mesh, value=self.n(t=float(t))
                )
            else:
                kwargs = {}
                if "t" in arguments:
                    kwargs["t"] = t
                if "x" in arguments:
                    kwargs["x"] = x

                self.value_fenics = self.n(**kwargs)

    def update_density(self, t):
        """Updates the density value (only if the density is a function of time only)

        Args:
            t (float): the time
        """
        if isinstance(self.n, (fem.Function, ufl.core.expr.Expr)):
            return

        if callable(self.n):
            arguments = self.n.__code__.co_varnames
            if isinstance(self.value_fenics, fem.Constant) and "t" in arguments:
                self.value_fenics.value = self.n(t=t)


def find_species_from_name(name: str, species: list):
    """Returns the correct species object from a list of species
    based on a string

    Args:
        name (str): the name of the species
        species (list): the list of species

    Returns:
        species (festim.Species): the species object with the correct name

    Raises:
        ValueError: if the species name is not found in the list of species

    """
    for spe in species:
        if spe.name == name:
            return spe
    raise ValueError(f"Species {name} not found in list of species")


class SpeciesChangeVar(Species):
    @property
    def concentration(self):
        return self._concentration

    @concentration.setter
    def concentration(self, value):
        self._concentration = value



import os

import pytest

import festim as F

mobile_H = F.Species("H")
mobile_D = F.Species("D")
surf_1 = F.SurfaceSubdomain(id=1)
surf_2 = F.SurfaceSubdomain(id=2)
vol_1 = F.VolumeSubdomain(id=1, material="test")
vol_2 = F.VolumeSubdomain(id=2, material="test")
results = "test.csv"

surface_flux = F.SurfaceFlux(field=mobile_H, surface=surf_1, filename=results)
average_vol = F.AverageVolume(mobile_H, volume=vol_1, filename=results)
tot_surf = F.TotalSurface(mobile_D, surface=surf_2, filename=results)
tot_vol = F.TotalVolume(mobile_D, volume=vol_2, filename=results)
min_vol = F.MinimumVolume(mobile_H, volume=vol_1, filename=results)
max_vol = F.MaximumVolume(mobile_D, volume=vol_1, filename=results)
min_surface = F.MinimumSurface(mobile_D, surface=surf_1, filename=results)
max_surface = F.MaximumSurface(mobile_H, surface=surf_2, filename=results)
avg_surface = F.AverageSurface(mobile_D, surface=surf_1, filename=results)
avg_vol = F.AverageVolume(mobile_H, volume=vol_2, filename=results)
surf_quant = F.SurfaceQuantity(mobile_H, surface=surf_1, filename=results)
vol_quant = F.VolumeQuantity(mobile_H, volume=vol_1, filename=results)


@pytest.mark.parametrize(
    "quantity, expected_title",
    [
        (surface_flux, "H flux surface 1"),
        (average_vol, "Average H volume 1"),
        (tot_surf, "Total D surface 2"),
        (tot_vol, "Total D volume 2"),
        (min_vol, "Minimum H volume 1"),
        (max_vol, "Maximum D volume 1"),
        (min_surface, "Minimum D surface 1"),
        (max_surface, "Maximum H surface 2"),
        (avg_surface, "Average D surface 1"),
        (avg_vol, "Average H volume 2"),
    ],
)
def test_title(quantity, expected_title, tmp_path):
    quantity.filename = os.path.join(tmp_path, "test.csv")
    quantity.value = 1

    assert quantity.title == expected_title



import os

import numpy as np
import pytest
import ufl
from dolfinx import fem

import festim as F

dummy_mat = F.Material(D_0=1, E_D=1, name="dummy")


def test_total_volume_export_compute():
    """Test that the total volume export computes the correct value"""

    # BUILD
    L = 4.0
    my_mesh = F.Mesh1D(np.linspace(0, L, 10000))
    dummy_volume = F.VolumeSubdomain1D(id=1, borders=[0, L], material=dummy_mat)
    temp, cell_meshtags = my_mesh.define_meshtags(
        surface_subdomains=[], volume_subdomains=[dummy_volume]
    )
    dx = ufl.Measure("dx", domain=my_mesh.mesh, subdomain_data=cell_meshtags)

    # give function to species
    V = fem.functionspace(my_mesh.mesh, ("Lagrange", 1))
    c = fem.Function(V)
    c.interpolate(lambda x: 2 * x[0] ** 2 + 1)

    my_species = F.Species("H")
    my_species.solution = c

    my_export = F.TotalVolume(
        field=my_species,
        volume=dummy_volume,
    )

    # RUN
    my_export.compute(dx=dx)

    # TEST
    # total = int[0,L] c dx = 2/3 * L^3 + L
    expected_value = 2 / 3 * L**3 + L
    computed_value = my_export.value

    assert np.isclose(computed_value, expected_value, rtol=1e-2)


@pytest.mark.parametrize("value", ["my_export.csv", "my_export.txt"])
def test_title_generation(tmp_path, value):
    """Test that the title is made to be written to the header in a csv or txt file"""
    my_export = F.TotalVolume(
        filename=os.path.join(tmp_path, f"{value}"),
        field=F.Species("TEST"),
        volume=F.VolumeSubdomain1D(id=35, borders=[0, 1], material=dummy_mat),
    )
    my_export.value = 2.0
    my_export.write(0)
    title = np.genfromtxt(my_export.filename, delimiter=",", max_rows=1, dtype=str)

    expected_title = "Total TEST volume 35"

    assert title[1] == expected_title



import festim as F
import numpy as np

import ufl
import dolfinx.fem as fem


# TODO add this to the festim package
class FluxFromSurfaceReaction(F.SurfaceFlux):
    def __init__(self, reaction: F.SurfaceReactionBC):
        super().__init__(
            F.Species(),  # just a dummy species here
            reaction.subdomain,
        )
        self.reaction = reaction.flux_bcs[0]

    def compute(self, ds):
        self.value = fem.assemble_scalar(
            fem.form(self.reaction.value_fenics * ds(self.surface.id))
        )
        self.data.append(self.value)


def test_2_isotopes_no_pressure():
    """
    Runs a simple 1D hydrogen transport problem with 2 isotopes
    and 3 surface reactions on the right boundary.

    Then checks that the fluxes of the isotopes are consistent with the surface reactions
    by computing the flux of H and D (from the gradient) and comparing it to the fluxes
    computed from the surface reactions.

    Example: H + D <-> HD, H + H <-> HH, D + D <-> DD
    -D grad(c_H) n = 2*kr*c_H*c_H + kr*c_H*c_D

    Also checks the mass balance between the left and right boundary
    """
    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh1D(vertices=np.linspace(0, 1, 1000))
    my_mat = F.Material(name="mat", D_0=1, E_D=0)
    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)

    my_model.subdomains = [vol, left, right]

    H = F.Species("H")
    D = F.Species("D")
    my_model.species = [H, D]

    my_model.temperature = 500

    surface_reaction_hd = F.SurfaceReactionBC(
        reactant=[H, D],
        gas_pressure=0,
        k_r0=0.01,
        E_kr=0,
        k_d0=0,
        E_kd=0,
        subdomain=right,
    )

    surface_reaction_hh = F.SurfaceReactionBC(
        reactant=[H, H],
        gas_pressure=0,
        k_r0=0.02,
        E_kr=0,
        k_d0=0,
        E_kd=0,
        subdomain=right,
    )

    surface_reaction_dd = F.SurfaceReactionBC(
        reactant=[D, D],
        gas_pressure=0,
        k_r0=0.03,
        E_kr=0,
        k_d0=0,
        E_kd=0,
        subdomain=right,
    )

    my_model.boundary_conditions = [
        F.DirichletBC(subdomain=left, value=2, species=H),
        F.DirichletBC(subdomain=left, value=3, species=D),
        surface_reaction_hd,
        surface_reaction_hh,
        surface_reaction_dd,
    ]

    H_flux_right = F.SurfaceFlux(H, right)
    H_flux_left = F.SurfaceFlux(H, left)
    D_flux_right = F.SurfaceFlux(D, right)
    D_flux_left = F.SurfaceFlux(D, left)
    HD_flux = FluxFromSurfaceReaction(surface_reaction_hd)
    HH_flux = FluxFromSurfaceReaction(surface_reaction_hh)
    DD_flux = FluxFromSurfaceReaction(surface_reaction_dd)
    my_model.exports = [
        H_flux_left,
        H_flux_right,
        D_flux_left,
        D_flux_right,
        HD_flux,
        HH_flux,
        DD_flux,
    ]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, final_time=5, transient=True)

    my_model.settings.stepsize = 0.1

    my_model.initialise()
    my_model.run()

    # TEST

    assert np.isclose(
        np.abs(H_flux_right.data[-1]),
        np.abs(H_flux_left.data[-1]),
        rtol=0.5e-2,
        atol=0.005,
    )
    assert np.isclose(
        np.abs(D_flux_right.data[-1]),
        np.abs(D_flux_left.data[-1]),
        rtol=0.5e-2,
        atol=0.005,
    )

    # check that H_flux_right == 2*HH_flux + HD_flux
    H_flux_from_gradient = -np.array(H_flux_right.data)
    H_flux_from_reac = 2 * np.array(HH_flux.data) + np.array(HD_flux.data)
    assert np.allclose(
        H_flux_from_gradient,
        H_flux_from_reac,
        rtol=0.5e-2,
        atol=0.005,
    )

    # check that D_flux_right == 2*DD_flux + HD_flux
    D_flux_from_gradient = -np.array(D_flux_right.data)
    D_flux_from_reac = 2 * np.array(DD_flux.data) + np.array(HD_flux.data)
    assert np.allclose(
        D_flux_from_gradient,
        D_flux_from_reac,
        rtol=0.5e-2,
        atol=0.005,
    )


def test_2_isotopes_with_pressure():
    """
    Runs a simple 1D hydrogen transport problem with 2 isotopes
    and 3 surface reactions on the right boundary.

    Then checks that the fluxes of the isotopes are consistent with the surface reactions
    by computing the flux of H and D (from the gradient) and comparing it to the fluxes
    computed from the surface reactions.

    Example: H + D <-> HD, H + H <-> HH, D + D <-> DD
    -D grad(c_H) n = 2*kr*c_H*c_H + kr*c_H*c_D - kd*P_H2

    Also checks the mass balance between the left and right boundary
    """
    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh1D(vertices=np.linspace(0, 1, 1000))
    my_mat = F.Material(name="mat", D_0=1, E_D=0)
    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)

    my_model.subdomains = [vol, left, right]

    H = F.Species("H")
    D = F.Species("D")
    my_model.species = [H, D]

    my_model.temperature = 500

    surface_reaction_hd = F.SurfaceReactionBC(
        reactant=[H, D],
        gas_pressure=0,
        k_r0=0.01,
        E_kr=0,
        k_d0=0,
        E_kd=0,
        subdomain=right,
    )

    surface_reaction_hh = F.SurfaceReactionBC(
        reactant=[H, H],
        gas_pressure=2,
        k_r0=0.02,
        E_kr=0,
        k_d0=0.1,
        E_kd=0,
        subdomain=right,
    )

    surface_reaction_dd = F.SurfaceReactionBC(
        reactant=[D, D],
        gas_pressure=0,
        k_r0=0.03,
        E_kr=0,
        k_d0=0,
        E_kd=0,
        subdomain=right,
    )

    my_model.boundary_conditions = [
        F.DirichletBC(subdomain=left, value=2, species=H),
        F.DirichletBC(subdomain=left, value=3, species=D),
        surface_reaction_hd,
        surface_reaction_hh,
        surface_reaction_dd,
    ]

    H_flux_right = F.SurfaceFlux(H, right)
    H_flux_left = F.SurfaceFlux(H, left)
    D_flux_right = F.SurfaceFlux(D, right)
    D_flux_left = F.SurfaceFlux(D, left)
    HD_flux = FluxFromSurfaceReaction(surface_reaction_hd)
    HH_flux = FluxFromSurfaceReaction(surface_reaction_hh)
    DD_flux = FluxFromSurfaceReaction(surface_reaction_dd)
    my_model.exports = [
        H_flux_left,
        H_flux_right,
        D_flux_left,
        D_flux_right,
        HD_flux,
        HH_flux,
        DD_flux,
    ]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, final_time=5, transient=True)

    my_model.settings.stepsize = 0.1

    my_model.initialise()
    my_model.run()

    # TEST

    assert np.isclose(
        np.abs(H_flux_right.data[-1]),
        np.abs(H_flux_left.data[-1]),
        rtol=0.5e-2,
        atol=0.005,
    )
    assert np.isclose(
        np.abs(D_flux_right.data[-1]),
        np.abs(D_flux_left.data[-1]),
        rtol=0.5e-2,
        atol=0.005,
    )

    # check that H_flux_right == 2*HH_flux + HD_flux
    H_flux_from_gradient = -np.array(H_flux_right.data)
    H_flux_from_reac = 2 * np.array(HH_flux.data) + np.array(HD_flux.data)
    assert np.allclose(
        H_flux_from_gradient,
        H_flux_from_reac,
        rtol=0.5e-2,
        atol=0.005,
    )

    # check that D_flux_right == 2*DD_flux + HD_flux
    D_flux_from_gradient = -np.array(D_flux_right.data)
    D_flux_from_reac = 2 * np.array(DD_flux.data) + np.array(HD_flux.data)
    assert np.allclose(
        D_flux_from_gradient,
        D_flux_from_reac,
        rtol=0.5e-2,
        atol=0.005,
    )


def test_pressure_varies_in_time():
    """
    Runs a problem with a surface reaction and a time-dependent pressure
    on the right boundary.

    Then checks that the flux is consistent with the surface reaction
    """
    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh1D(vertices=np.linspace(0, 1, 1000))
    my_mat = F.Material(name="mat", D_0=1, E_D=0)
    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)

    my_model.subdomains = [vol, left, right]

    H = F.Species("H")
    my_model.species = [H]

    my_model.temperature = 500

    t_pressure = 2
    pressure = 2
    k_d = 2

    surface_reaction_hh = F.SurfaceReactionBC(
        reactant=[H, H],
        gas_pressure=lambda t: ufl.conditional(ufl.gt(t, t_pressure), pressure, 0),
        k_r0=0,
        E_kr=0,
        k_d0=k_d,
        E_kd=0,
        subdomain=right,
    )

    my_model.boundary_conditions = [surface_reaction_hh]

    H_flux_right = F.SurfaceFlux(H, right)
    my_model.exports = [H_flux_right]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, final_time=5, transient=True)

    my_model.settings.stepsize = 0.1

    my_model.initialise()
    my_model.run()

    flux_as_array = np.array(H_flux_right.data)
    time_as_array = np.array(H_flux_right.t)

    expected_flux_before_pressure = 0
    computed_flux_before_pressure = flux_as_array[time_as_array <= t_pressure]
    assert np.allclose(computed_flux_before_pressure, expected_flux_before_pressure)

    expected_flux_after_pressure = -2 * k_d * pressure
    computed_flux_after_pressure = flux_as_array[time_as_array > t_pressure]
    print(computed_flux_after_pressure)
    print(expected_flux_after_pressure)
    assert np.allclose(
        computed_flux_after_pressure, expected_flux_after_pressure, rtol=1e-2
    )



import os

from petsc4py import PETSc

import numpy as np
from dolfinx.fem import Constant
from ufl import exp

import festim as F


def relative_error_computed_to_analytical(
    D, permeability, computed_flux, L, times, P_up
):
    n_array = np.arange(1, 10000)[:, np.newaxis]
    summation = np.sum(
        (-1) ** n_array * np.exp(-((np.pi * n_array) ** 2) * float(D) / L**2 * times),
        axis=0,
    )
    analytical_flux = P_up**0.5 * permeability / L * (2 * summation + 1)

    # post processing
    analytical_flux = np.abs(analytical_flux)
    indices = np.where(analytical_flux > 0.1 * np.max(analytical_flux))
    analytical_flux = analytical_flux[indices]
    computed_flux = computed_flux[indices]

    # evaulate relative error compared to analytical solution
    relative_error = np.abs((computed_flux - analytical_flux) / analytical_flux)
    error = relative_error.mean()

    return error


def test_permeation_problem(mesh_size=1001):
    """Test running a problem with a mobile species permeating through a 1D
    domain, checks that the computed permeation flux matches the analytical
    solution"""

    # festim model
    L = 3e-04
    vertices = np.linspace(0, L, num=mesh_size)

    my_mesh = F.Mesh1D(vertices)

    my_model = F.HydrogenTransportProblem()
    my_model.mesh = my_mesh

    my_mat = F.Material(D_0=1.9e-7, E_D=0.2, name="my_mat")
    my_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, L], material=my_mat)
    left_surface = F.SurfaceSubdomain1D(id=1, x=0)
    right_surface = F.SurfaceSubdomain1D(id=2, x=L)
    my_model.subdomains = [my_subdomain, left_surface, right_surface]

    mobile_H = F.Species("H")
    my_model.species = [mobile_H]

    my_model.temperature = 500

    my_model.boundary_conditions = [
        F.DirichletBC(subdomain=right_surface, value=0, species="H"),
        F.SievertsBC(
            subdomain=left_surface,
            S_0=4.02e21,
            E_S=1.04,
            pressure=100,
            species="H",
        ),
    ]
    outgassing_flux = F.SurfaceFlux(
        filename="outgassing_flux.txt",
        field=mobile_H,
        surface=right_surface,
    )
    my_model.exports = [
        F.XDMFExport("mobile_concentration.xdmf", field=mobile_H),
        outgassing_flux,
    ]

    my_model.settings = F.Settings(
        atol=1e10,
        rtol=1e-10,
        max_iterations=30,
        final_time=50,
    )

    my_model.settings.stepsize = F.Stepsize(initial_value=1 / 20)

    my_model.initialise()

    my_model.solver.convergence_criterion = "incremental"
    ksp = my_model.solver.krylov_solver
    opts = PETSc.Options()
    option_prefix = ksp.getOptionsPrefix()
    opts[f"{option_prefix}ksp_type"] = "cg"
    opts[f"{option_prefix}pc_type"] = "gamg"
    opts[f"{option_prefix}pc_factor_mat_solver_type"] = "mumps"
    ksp.setFromOptions()

    my_model.run()

    times = outgassing_flux.t
    flux_values = outgassing_flux.data

    # -------------------------- analytical solution -------------------------------------

    D = my_mat.get_diffusion_coefficient(my_mesh.mesh, my_model.temperature)

    S_0 = float(my_model.boundary_conditions[-1].S_0)
    E_S = float(my_model.boundary_conditions[-1].E_S)
    P_up = float(my_model.boundary_conditions[-1].pressure)
    S = S_0 * exp(-E_S / F.k_B / float(my_model.temperature))
    permeability = float(D) * S
    times = np.array(times)
    flux_values = np.array(np.abs(flux_values))

    error = relative_error_computed_to_analytical(
        D, permeability, flux_values, L, times, P_up
    )

    assert error < 0.01


def test_permeation_problem_multi_volume(tmp_path):
    """Same permeation problem as above but with 4 volume subdomains instead
    of 1"""

    L = 3e-04
    vertices = np.linspace(0, L, num=1001)

    my_mesh = F.Mesh1D(vertices)

    my_model = F.HydrogenTransportProblem()
    my_model.mesh = my_mesh

    my_mat = F.Material(D_0=1.9e-7, E_D=0.2, name="my_mat")
    my_subdomain_1 = F.VolumeSubdomain1D(id=1, borders=[0, L / 4], material=my_mat)
    my_subdomain_2 = F.VolumeSubdomain1D(id=2, borders=[L / 4, L / 2], material=my_mat)
    my_subdomain_3 = F.VolumeSubdomain1D(
        id=3, borders=[L / 2, 3 * L / 4], material=my_mat
    )
    my_subdomain_4 = F.VolumeSubdomain1D(id=4, borders=[3 * L / 4, L], material=my_mat)
    left_surface = F.SurfaceSubdomain1D(id=1, x=0)
    right_surface = F.SurfaceSubdomain1D(id=2, x=L)
    my_model.subdomains = [
        my_subdomain_1,
        my_subdomain_2,
        my_subdomain_3,
        my_subdomain_4,
        left_surface,
        right_surface,
    ]

    mobile_H = F.Species("H")
    my_model.species = [mobile_H]

    temperature = Constant(my_mesh.mesh, 500.0)
    my_model.temperature = temperature

    my_model.boundary_conditions = [
        F.DirichletBC(subdomain=right_surface, value=0, species="H"),
        F.SievertsBC(
            subdomain=left_surface,
            S_0=4.02e21,
            E_S=1.04,
            pressure=100,
            species="H",
        ),
    ]
    outgassing_flux = F.SurfaceFlux(
        filename=os.path.join(tmp_path, "outgassing_flux.csv"),
        field=mobile_H,
        surface=right_surface,
    )
    my_model.exports = [
        F.VTXSpeciesExport(
            os.path.join(tmp_path, "mobile_concentration_h.bp"), field=mobile_H
        ),
        outgassing_flux,
    ]

    my_model.settings = F.Settings(
        atol=1e10,
        rtol=1e-10,
        max_iterations=30,
        final_time=50,
    )

    my_model.settings.stepsize = F.Stepsize(initial_value=1 / 20)

    my_model.initialise()

    my_model.solver.convergence_criterion = "incremental"
    ksp = my_model.solver.krylov_solver
    opts = PETSc.Options()
    option_prefix = ksp.getOptionsPrefix()
    opts[f"{option_prefix}ksp_type"] = "cg"
    opts[f"{option_prefix}pc_type"] = "gamg"
    opts[f"{option_prefix}pc_factor_mat_solver_type"] = "mumps"
    ksp.setFromOptions()

    my_model.run()

    times = outgassing_flux.t
    flux_values = outgassing_flux.data

    # ---------------------- analytical solution -----------------------------
    D = my_mat.get_diffusion_coefficient(my_mesh.mesh, temperature)

    S_0 = float(my_model.boundary_conditions[-1].S_0)
    E_S = float(my_model.boundary_conditions[-1].E_S)
    P_up = float(my_model.boundary_conditions[-1].pressure)
    S = S_0 * exp(-E_S / F.k_B / float(temperature))
    permeability = float(D) * S
    times = np.array(times)
    flux_values = np.array(np.abs(flux_values))

    error = relative_error_computed_to_analytical(
        D, permeability, flux_values, L, times, P_up
    )

    assert error < 0.01
